import importlib
import os
import time

import pytest

import salt.config
import salt.exceptions
import salt.fileserver.gitfs
import salt.utils.gitfs
from salt.exceptions import FileserverConfigError
from tests.support.helpers import patched_environ
from tests.support.mock import MagicMock, patch

try:
    HAS_PYGIT2 = (
        salt.utils.gitfs.PYGIT2_VERSION
        and salt.utils.gitfs.PYGIT2_VERSION >= salt.utils.gitfs.PYGIT2_MINVER
        and salt.utils.gitfs.LIBGIT2_VERSION
        and salt.utils.gitfs.LIBGIT2_VERSION >= salt.utils.gitfs.LIBGIT2_MINVER
    )
except AttributeError:
    HAS_PYGIT2 = False


if HAS_PYGIT2:
    import pygit2

    try:
        from pygit2.enums import ObjectType

        HAS_PYGIT2_ENUMS = True

    except ModuleNotFoundError:
        HAS_PYGIT2_ENUMS = False


@pytest.fixture
def minion_opts(tmp_path):
    """
    Default minion configuration with relative temporary paths to not require root permissions.
    """
    root_dir = tmp_path / "minion"
    opts = salt.config.DEFAULT_MINION_OPTS.copy()
    opts["__role"] = "minion"
    opts["root_dir"] = str(root_dir)
    for name in ("cachedir", "pki_dir", "sock_dir", "conf_dir"):
        dirpath = root_dir / name
        dirpath.mkdir(parents=True)
        opts[name] = str(dirpath)
    opts["log_file"] = "logs/minion.log"
    return opts


@pytest.mark.parametrize(
    "role_name,role_class",
    (
        ("gitfs", salt.utils.gitfs.GitFS),
        ("git_pillar", salt.utils.gitfs.GitPillar),
        ("winrepo", salt.utils.gitfs.WinRepo),
    ),
)
def test_provider_case_insensitive_gitfs_provider(minion_opts, role_name, role_class):
    """
    Ensure that both lowercase and non-lowercase values are supported
    """
    provider = "GitPython"
    key = f"{role_name}_provider"
    with patch.object(role_class, "verify_gitpython", MagicMock(return_value=True)):
        with patch.object(role_class, "verify_pygit2", MagicMock(return_value=False)):
            args = [minion_opts, {}]
            kwargs = {"init_remotes": False}
            if role_name == "winrepo":
                kwargs["cache_root"] = "/tmp/winrepo-dir"
            with patch.dict(minion_opts, {key: provider}):
                # Try to create an instance with uppercase letters in
                # provider name. If it fails then a
                # FileserverConfigError will be raised, so no assert is
                # necessary.
                role_class(*args, **kwargs)
            # Now try to instantiate an instance with all lowercase
            # letters. Again, no need for an assert here.
            role_class(*args, **kwargs)


@pytest.mark.parametrize(
    "role_name,role_class",
    (
        ("gitfs", salt.utils.gitfs.GitFS),
        ("git_pillar", salt.utils.gitfs.GitPillar),
        ("winrepo", salt.utils.gitfs.WinRepo),
    ),
)
def test_valid_provider_gitfs_provider(minion_opts, role_name, role_class):
    """
    Ensure that an invalid provider is not accepted, raising a
    FileserverConfigError.
    """

    def _get_mock(verify, provider):
        """
        Return a MagicMock with the desired return value
        """
        return MagicMock(return_value=verify.endswith(provider))

    key = f"{role_name}_provider"
    for provider in salt.utils.gitfs.GIT_PROVIDERS:
        verify = "verify_gitpython"
        mock1 = _get_mock(verify, provider)
        with patch.object(role_class, verify, mock1):
            verify = "verify_pygit2"
            mock2 = _get_mock(verify, provider)
            with patch.object(role_class, verify, mock2):
                args = [minion_opts, {}]
                kwargs = {"init_remotes": False}
                if role_name == "winrepo":
                    kwargs["cache_root"] = "/tmp/winrepo-dir"
                with patch.dict(minion_opts, {key: provider}):
                    role_class(*args, **kwargs)
                with patch.dict(minion_opts, {key: "foo"}):
                    # Set the provider name to a known invalid provider
                    # and make sure it raises an exception.
                    with pytest.raises(FileserverConfigError):
                        role_class(*args, **kwargs)


@pytest.fixture
def _prepare_remote_repository_pygit2(tmp_path):
    remote = os.path.join(tmp_path, "pygit2-repo")
    filecontent = "This is an empty README file"
    filename = "README"
    signature = pygit2.Signature(
        "Dummy Commiter", "dummy@dummy.com", int(time.time()), 0
    )
    repository = pygit2.init_repository(remote, False)
    builder = repository.TreeBuilder()
    tree = builder.write()
    commit = repository.create_commit(
        "HEAD", signature, signature, "Create master branch", tree, []
    )
    repository.create_reference("refs/tags/simple_tag", commit)
    with salt.utils.files.fopen(
        os.path.join(repository.workdir, filename), "w"
    ) as file:
        file.write(filecontent)
    blob = repository.create_blob_fromworkdir(filename)
    builder = repository.TreeBuilder()
    builder.insert(filename, blob, pygit2.GIT_FILEMODE_BLOB)
    tree = builder.write()
    repository.index.read()
    repository.index.add(filename)
    repository.index.write()
    commit = repository.create_commit(
        "HEAD",
        signature,
        signature,
        "Added a README",
        tree,
        [repository.head.target],
    )
    if HAS_PYGIT2_ENUMS:
        repository.create_tag(
            "annotated_tag", commit, ObjectType.COMMIT, signature, "some message"
        )
    else:
        repository.create_tag(
            "annotated_tag", commit, pygit2.GIT_OBJ_COMMIT, signature, "some message"
        )
    return remote


@pytest.fixture
def _prepare_provider(tmp_path, minion_opts, _prepare_remote_repository_pygit2):
    cache = tmp_path / "pygit2-repo-cache"
    minion_opts.update(
        {
            "cachedir": str(cache),
            "gitfs_disable_saltenv_mapping": False,
            "gitfs_base": "master",
            "gitfs_insecure_auth": False,
            "gitfs_mountpoint": "",
            "gitfs_passphrase": "",
            "gitfs_password": "",
            "gitfs_privkey": "",
            "gitfs_provider": "pygit2",
            "gitfs_pubkey": "",
            "gitfs_ref_types": ["branch", "tag", "sha"],
            "gitfs_refspecs": [
                "+refs/heads/*:refs/remotes/origin/*",
                "+refs/tags/*:refs/tags/*",
            ],
            "gitfs_root": "",
            "gitfs_saltenv_blacklist": [],
            "gitfs_saltenv_whitelist": [],
            "gitfs_ssl_verify": True,
            "gitfs_update_interval": 3,
            "gitfs_user": "",
            "verified_gitfs_provider": "pygit2",
        }
    )
    per_remote_defaults = {
        "base": "master",
        "disable_saltenv_mapping": False,
        "insecure_auth": False,
        "ref_types": ["branch", "tag", "sha"],
        "passphrase": "",
        "mountpoint": "",
        "password": "",
        "privkey": "",
        "pubkey": "",
        "refspecs": [
            "+refs/heads/*:refs/remotes/origin/*",
            "+refs/tags/*:refs/tags/*",
        ],
        "root": "",
        "saltenv_blacklist": [],
        "saltenv_whitelist": [],
        "ssl_verify": True,
        "update_interval": 60,
        "user": "",
    }
    per_remote_only = ("all_saltenvs", "name", "saltenv")
    override_params = tuple(per_remote_defaults)
    cache_root = cache / "gitfs"
    role = "gitfs"
    provider = salt.utils.gitfs.Pygit2(
        minion_opts,
        _prepare_remote_repository_pygit2,
        per_remote_defaults,
        per_remote_only,
        override_params,
        str(cache_root),
        role,
    )
    return provider


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
@pytest.mark.skip_on_windows(
    reason="Skip Pygit2 on windows, due to pygit2 access error on windows"
)
def test_checkout_pygit2(_prepare_provider):
    provider = _prepare_provider
    provider.remotecallbacks = None
    provider.credentials = None
    provider.init_remote()
    provider.fetch()
    provider.branch = "master"
    assert provider.get_cachedir() in provider.checkout()
    provider.branch = "simple_tag"
    assert provider.get_cachedir() in provider.checkout()
    provider.branch = "annotated_tag"
    assert provider.get_cachedir() in provider.checkout()
    provider.branch = "does_not_exist"
    assert provider.checkout() is None


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
@pytest.mark.skip_on_windows(
    reason="Skip Pygit2 on windows, due to pygit2 access error on windows"
)
def test_checkout_pygit2_with_home_env_unset(_prepare_provider):
    provider = _prepare_provider
    provider.remotecallbacks = None
    provider.credentials = None
    with patched_environ(__cleanup__=["HOME"]):
        assert "HOME" not in os.environ
        importlib.reload(salt.utils.gitfs)
        assert "HOME" in os.environ


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
@pytest.mark.skip_on_windows(
    reason="Skip Pygit2 on windows, due to pygit2 access error on windows"
)
@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
@pytest.mark.skip_on_windows(
    reason="Skip Pygit2 on windows, due to pygit2 access error on windows"
)
def test_get_cachedir_basename_pygit2(_prepare_provider):
    assert "_" == _prepare_provider.get_cache_basename()


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
def test_find_file(tmp_path):
    opts = {
        "cachedir": f"{tmp_path / 'cache'}",
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_pubkey": "",
        "gitfs_privkey": "",
        "gitfs_passphrase": "",
        "gitfs_insecure_auth": False,
        "gitfs_refspecs": salt.config._DFLT_REFSPECS,
        "gitfs_ssl_verify": True,
        "gitfs_branch": "master",
        "gitfs_base": "master",
        "gitfs_root": "",
        "gitfs_env": "",
        "gitfs_fallback": "",
    }
    remotes = []

    gitfs = salt.utils.gitfs.GitFS(opts, remotes)
    assert gitfs.find_file("asdf") == {"path": "", "rel": ""}


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
def test_find_file_bad_path(tmp_path):
    opts = {
        "cachedir": f"{tmp_path / 'cache'}",
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_pubkey": "",
        "gitfs_privkey": "",
        "gitfs_passphrase": "",
        "gitfs_insecure_auth": False,
        "gitfs_refspecs": salt.config._DFLT_REFSPECS,
        "gitfs_ssl_verify": True,
        "gitfs_branch": "master",
        "gitfs_base": "master",
        "gitfs_root": "",
        "gitfs_env": "",
        "gitfs_fallback": "",
    }
    remotes = []

    gitfs = salt.utils.gitfs.GitFS(opts, remotes)
    with pytest.raises(salt.exceptions.SaltValidationError):
        gitfs.find_file("sdf/../../../asdf")


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
def test_find_file_bad_env(tmp_path):
    opts = {
        "cachedir": f"{tmp_path / 'cache'}",
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_pubkey": "",
        "gitfs_privkey": "",
        "gitfs_passphrase": "",
        "gitfs_insecure_auth": False,
        "gitfs_refspecs": salt.config._DFLT_REFSPECS,
        "gitfs_ssl_verify": True,
        "gitfs_branch": "master",
        "gitfs_base": "master",
        "gitfs_root": "",
        "gitfs_env": "",
        "gitfs_fallback": "",
    }
    remotes = []

    gitfs = salt.utils.gitfs.GitFS(opts, remotes)
    with pytest.raises(salt.exceptions.SaltValidationError):
        gitfs.find_file("asdf", tgt_env="asd/../../../sdf")


@pytest.mark.parametrize(
    "remote,valid",
    [
        ("git@github.com:/saltstack/salt", True),
        ("git@github.com:saltstack/salt", True),
        ("git@github.com/saltstack/salt", False),
        ("ssh://git@github.com/saltstack/salt.git", True),
        ("ssh://git@github.com:22/saltstack/salt.git", True),
        ("https://github.com/salttack/salt.git", True),
        ("https://github.com/\nsaltstack/salt.git", False),
        ("https://git:mypassword@github.com/saltstack/salt.git", True),
        ("file:///srv/git/salt.git", True),
    ],
)
def test_remote_validation(remote, valid):
    assert salt.utils.gitfs.GitFS.validate_remote(remote) is valid


@pytest.mark.parametrize(
    "remote,result",
    [
        ("git@github.com:/saltstack/salt", "ssh://git@github.com/saltstack/salt"),
        ("git@github.com:saltstack/salt", "ssh://git@github.com/saltstack/salt"),
        (
            "ssh://git@github.com/saltstack/salt.git",
            "ssh://git@github.com/saltstack/salt.git",
        ),
        (
            "ssh://git@github.com:22/saltstack/salt.git",
            "ssh://git@github.com:22/saltstack/salt.git",
        ),
        (
            "https://github.com/salttack/salt.git",
            "https://github.com/salttack/salt.git",
        ),
        (
            "https://git:mypassword@github.com/saltstack/salt.git",
            "https://git:mypassword@github.com/saltstack/salt.git",
        ),
        ("file:///srv/git/salt.git", "file:///srv/git/salt.git"),
    ],
)
def test_remote_to_url(remote, result):
    assert salt.utils.gitfs.GitFS.remote_to_url(remote) == result


@pytest.mark.skipif(not HAS_PYGIT2, reason="This host lacks proper pygit2 support")
def test_find_file_subdir(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "refs").mkdir()
    (root / "refs" / "base").mkdir()
    opts = {
        "cachedir": f"{tmp_path / 'cache'}",
        "gitfs_user": "",
        "gitfs_password": "",
        "gitfs_pubkey": "",
        "gitfs_privkey": "",
        "gitfs_passphrase": "",
        "gitfs_insecure_auth": False,
        "gitfs_refspecs": salt.config._DFLT_REFSPECS,
        "gitfs_ssl_verify": True,
        "gitfs_branch": "master",
        "gitfs_base": "master",
        "gitfs_root": "",
        "gitfs_env": "",
        "gitfs_fallback": "",
    }
    remotes = []
    gitfs = salt.utils.gitfs.GitFS(opts, remotes)
    gitfs.cache_root = str(root)
    ret = gitfs.find_file("foo/init.sls")
    assert ret == {"path": "", "rel": ""}



import os

import pytest

from salt.runners.winrepo import GLOBAL_ONLY, PER_REMOTE_ONLY, PER_REMOTE_OVERRIDES
from salt.utils.gitfs import GitPython, Pygit2, WinRepo
from salt.utils.immutabletypes import ImmutableDict, ImmutableList
from salt.utils.platform import get_machine_identifier as _get_machine_identifier

pytestmark = [
    pytest.mark.slow_test,
]


try:
    import git  # pylint: disable=unused-import

    HAS_GITPYTHON = True
except ImportError:
    HAS_GITPYTHON = False


try:
    import pygit2  # pylint: disable=unused-import

    HAS_PYGIT2 = True
except ImportError:
    HAS_PYGIT2 = False


skipif_no_gitpython = pytest.mark.skipif(not HAS_GITPYTHON, reason="Missing gitpython")
skipif_no_pygit2 = pytest.mark.skipif(not HAS_PYGIT2, reason="Missing pygit2")


@pytest.fixture
def winrepo_opts(salt_factories, tmp_path):
    config_defaults = {"cachedir": str(tmp_path)}
    factory = salt_factories.salt_master_daemon(
        "winrepo-functional-master", defaults=config_defaults
    )
    config_defaults = dict(factory.config)
    for key, item in config_defaults.items():
        if isinstance(item, ImmutableDict):
            config_defaults[key] = dict(item)
        elif isinstance(item, ImmutableList):
            config_defaults[key] = list(item)
    return config_defaults


@pytest.fixture
def gitpython_winrepo_opts(winrepo_opts):
    winrepo_opts["verified_winrepo_provider"] = "gitpython"
    return winrepo_opts


@pytest.fixture
def pygit2_winrepo_opts(winrepo_opts):
    winrepo_opts["verified_winrepo_provider"] = "pygit2"
    return winrepo_opts


def _get_winrepo(opts, *remotes):
    return WinRepo(
        opts,
        remotes,
        per_remote_overrides=PER_REMOTE_OVERRIDES,
        per_remote_only=PER_REMOTE_ONLY,
        global_only=GLOBAL_ONLY,
    )


@skipif_no_gitpython
def test_gitpython_winrepo_provider(gitpython_winrepo_opts):
    w = _get_winrepo(
        gitpython_winrepo_opts,
        "https://github.com/saltstack/salt-test-pillar-gitfs.git",
    )
    assert len(w.remotes) == 1
    assert w.provider == "gitpython"
    assert isinstance(w.remotes[0], GitPython)


@skipif_no_pygit2
def test_pygit2_winrepo_provider(pygit2_winrepo_opts):
    w = _get_winrepo(
        pygit2_winrepo_opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git"
    )
    assert len(w.remotes) == 1
    assert w.provider == "pygit2"
    assert isinstance(w.remotes[0], Pygit2)


def _test_winrepo_simple(opts):
    w = _get_winrepo(opts, "https://github.com/saltstack/salt-test-pillar-gitfs.git")
    assert len(w.remotes) == 1
    w.checkout()
    repo = w.remotes[0]
    files = set(os.listdir(repo.get_cachedir()))
    for f in (".gitignore", "README.md", "file.sls", "top.sls"):
        assert f in files


@skipif_no_gitpython
def test_gitpython_winrepo_simple(gitpython_winrepo_opts):
    _test_winrepo_simple(gitpython_winrepo_opts)


@skipif_no_pygit2
def test_pygit2_winrepo_simple(pygit2_winrepo_opts):
    _test_winrepo_simple(pygit2_winrepo_opts)


def _test_remote_map(opts):
    p = _get_winrepo(
        opts,
        "https://github.com/saltstack/salt-test-pillar-gitfs.git",
    )
    p.fetch_remotes()
    assert len(p.remotes) == 1
    assert os.path.isfile(os.path.join(opts["cachedir"], "winrepo", "remote_map.txt"))


@skipif_no_gitpython
def test_gitpython_remote_map(gitpython_winrepo_opts):
    _test_remote_map(gitpython_winrepo_opts)


@skipif_no_pygit2
def test_pygit2_remote_map(pygit2_winrepo_opts):
    _test_remote_map(pygit2_winrepo_opts)


def _test_lock(opts):
    mach_id = _get_machine_identifier().get("machine_id", "no_machine_id_available")
    w = _get_winrepo(
        opts,
        "https://github.com/saltstack/salt-test-pillar-gitfs.git",
    )
    w.fetch_remotes()
    assert len(w.remotes) == 1
    repo = w.remotes[0]
    assert repo.get_salt_working_dir() in repo._get_lock_file()
    assert repo.lock() == (
        [
            (
                f"Set update lock for winrepo remote 'https://github.com/saltstack/salt-test-pillar-gitfs.git' on machine_id '{mach_id}'"
            )
        ],
        [],
    )
    assert os.path.isfile(repo._get_lock_file())
    assert repo.clear_lock() == (
        [
            (
                f"Removed update lock for winrepo remote 'https://github.com/saltstack/salt-test-pillar-gitfs.git' on machine_id '{mach_id}'"
            )
        ],
        [],
    )
    assert not os.path.isfile(repo._get_lock_file())


@skipif_no_gitpython
def test_gitpython_lock(gitpython_winrepo_opts):
    _test_lock(gitpython_winrepo_opts)


@skipif_no_pygit2
def test_pygit2_lock(pygit2_winrepo_opts):
    _test_lock(pygit2_winrepo_opts)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.win_ip
"""

import pytest

import salt.modules.win_ip as win_ip
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {win_ip: {}}


@pytest.fixture
def ethernet_config():
    return (
        'Configuration for interface "Ethernet"\n'
        "DHCP enabled: Yes\n"
        "IP Address: 1.2.3.74\n"
        "Subnet Prefix: 1.2.3.0/24 (mask 255.255.255.0)\n"
        "Default Gateway: 1.2.3.1\n"
        "Gateway Metric: 0\n"
        "InterfaceMetric: 20\n"
        "DNS servers configured through DHCP: 1.2.3.4\n"
        "Register with which suffix: Primary only\n"
        "WINS servers configured through DHCP: None\n"
    )


@pytest.fixture
def ethernet_enable():
    return "Ethernet\nType: Dedicated\nAdministrative state: Enabled\nConnect state: Connected"


# 'raw_interface_configs' function tests: 1


def test_raw_interface_configs(ethernet_config):
    """
    Test if it return raw configs for all interfaces.
    """
    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.raw_interface_configs() == ethernet_config


# 'get_all_interfaces' function tests: 1


def test_get_all_interfaces(ethernet_config):
    """
    Test if it return configs for all interfaces.
    """
    ret = {
        "Ethernet": {
            "DHCP enabled": "Yes",
            "DNS servers configured through DHCP": ["1.2.3.4"],
            "Default Gateway": "1.2.3.1",
            "Gateway Metric": "0",
            "InterfaceMetric": "20",
            "Register with which suffix": "Primary only",
            "WINS servers configured through DHCP": ["None"],
            "ip_addrs": [
                {
                    "IP Address": "1.2.3.74",
                    "Netmask": "255.255.255.0",
                    "Subnet": "1.2.3.0/24",
                }
            ],
        }
    }

    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.get_all_interfaces() == ret


# 'get_interface' function tests: 1


def test_get_interface(ethernet_config):
    """
    Test if it return the configuration of a network interface.
    """
    ret = {
        "DHCP enabled": "Yes",
        "DNS servers configured through DHCP": ["1.2.3.4"],
        "Default Gateway": "1.2.3.1",
        "Gateway Metric": "0",
        "InterfaceMetric": "20",
        "Register with which suffix": "Primary only",
        "WINS servers configured through DHCP": ["None"],
        "ip_addrs": [
            {
                "IP Address": "1.2.3.74",
                "Netmask": "255.255.255.0",
                "Subnet": "1.2.3.0/24",
            }
        ],
    }

    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.get_interface("Ethernet") == ret


# 'is_enabled' function tests: 1


def test_is_enabled(ethernet_enable):
    """
    Test if it returns `True` if interface is enabled, otherwise `False`.
    """
    mock_cmd = MagicMock(side_effect=[ethernet_enable, ""])
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.is_enabled("Ethernet")
        pytest.raises(CommandExecutionError, win_ip.is_enabled, "Ethernet")


# 'is_disabled' function tests: 1


def test_is_disabled(ethernet_enable):
    """
    Test if it returns `True` if interface is disabled, otherwise `False`.
    """
    mock_cmd = MagicMock(return_value=ethernet_enable)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert not win_ip.is_disabled("Ethernet")


# 'enable' function tests: 1


def test_enable():
    """
    Test if it enable an interface.
    """
    # Test with enabled interface
    with patch.object(win_ip, "is_enabled", return_value=True):
        assert win_ip.enable("Ethernet")

    mock_cmd = MagicMock()
    with patch.object(win_ip, "is_enabled", side_effect=[False, True]), patch.dict(
        win_ip.__salt__, {"cmd.run": mock_cmd}
    ):
        assert win_ip.enable("Ethernet")

    mock_cmd.assert_called_once_with(
        [
            "netsh",
            "interface",
            "set",
            "interface",
            "name=Ethernet",
            "admin=ENABLED",
        ],
        python_shell=False,
    )


# 'disable' function tests: 1


def test_disable():
    """
    Test if it disable an interface.
    """
    with patch.object(win_ip, "is_disabled", return_value=True):
        assert win_ip.disable("Ethernet")

    mock_cmd = MagicMock()
    with patch.object(win_ip, "is_disabled", side_effect=[False, True]), patch.dict(
        win_ip.__salt__, {"cmd.run": mock_cmd}
    ):
        assert win_ip.disable("Ethernet")

    mock_cmd.assert_called_once_with(
        [
            "netsh",
            "interface",
            "set",
            "interface",
            "name=Ethernet",
            "admin=DISABLED",
        ],
        python_shell=False,
    )


# 'get_subnet_length' function tests: 1


def test_get_subnet_length():
    """
    Test if it disable an interface.
    """
    assert win_ip.get_subnet_length("255.255.255.0") == 24
    pytest.raises(SaltInvocationError, win_ip.get_subnet_length, "255.255.0")


# 'set_static_ip' function tests: 1


@pytest.mark.slow_test
def test_set_static_ip(ethernet_config):
    """
    Test if it set static IP configuration on a Windows NIC.
    """
    pytest.raises(
        SaltInvocationError,
        win_ip.set_static_ip,
        "Local Area Connection",
        "10.1.2/24",
    )

    mock_cmd = MagicMock(return_value=ethernet_config)
    mock_all = MagicMock(return_value={"retcode": 1, "stderr": "Error"})
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd, "cmd.run_all": mock_all}):
        pytest.raises(
            CommandExecutionError,
            win_ip.set_static_ip,
            "Ethernet",
            "1.2.3.74/24",
            append=True,
        )
        pytest.raises(
            CommandExecutionError, win_ip.set_static_ip, "Ethernet", "1.2.3.74/24"
        )

    mock_all = MagicMock(return_value={"retcode": 0})
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd, "cmd.run_all": mock_all}):
        assert win_ip.set_static_ip("Local Area Connection", "1.2.3.74/24") == {}
        assert win_ip.set_static_ip("Ethernet", "1.2.3.74/24") == {
            "Address Info": {
                "IP Address": "1.2.3.74",
                "Netmask": "255.255.255.0",
                "Subnet": "1.2.3.0/24",
            }
        }


# 'set_dhcp_ip' function tests: 1


def test_set_dhcp_ip(ethernet_config):
    """
    Test if it set Windows NIC to get IP from DHCP.
    """
    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.set_dhcp_ip("Ethernet") == {
            "DHCP enabled": "Yes",
            "Interface": "Ethernet",
        }


# 'set_static_dns' function tests: 1


def test_set_static_dns():
    """
    Test if it set static DNS configuration on a Windows NIC.
    """
    mock_cmd = MagicMock()
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.set_static_dns("Ethernet", "192.168.1.252", "192.168.1.253") == {
            "DNS Server": ("192.168.1.252", "192.168.1.253"),
            "Interface": "Ethernet",
        }
        mock_cmd.assert_has_calls(
            [
                call(
                    [
                        "netsh",
                        "interface",
                        "ip",
                        "set",
                        "dns",
                        "name=Ethernet",
                        "source=static",
                        "address=192.168.1.252",
                        "register=primary",
                    ],
                    python_shell=False,
                ),
                call(
                    [
                        "netsh",
                        "interface",
                        "ip",
                        "add",
                        "dns",
                        "name=Ethernet",
                        "address=192.168.1.253",
                        "index=2",
                    ],
                    python_shell=False,
                ),
            ]
        )


def test_set_static_dns_clear():
    """
    Test if it set static DNS configuration on a Windows NIC.
    """
    mock_cmd = MagicMock()
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.set_static_dns("Ethernet", []) == {
            "DNS Server": [],
            "Interface": "Ethernet",
        }
        mock_cmd.assert_called_once_with(
            [
                "netsh",
                "interface",
                "ip",
                "set",
                "dns",
                "name=Ethernet",
                "source=static",
                "address=none",
            ],
            python_shell=False,
        )


def test_set_static_dns_no_action():
    """
    Test if it set static DNS configuration on a Windows NIC.
    """
    # Test passing nothing
    assert win_ip.set_static_dns("Ethernet") == {
        "DNS Server": "No Changes",
        "Interface": "Ethernet",
    }
    # Test passing None
    assert win_ip.set_static_dns("Ethernet", None) == {
        "DNS Server": "No Changes",
        "Interface": "Ethernet",
    }

    # Test passing string None
    assert win_ip.set_static_dns("Ethernet", "None") == {
        "DNS Server": "No Changes",
        "Interface": "Ethernet",
    }


# 'set_dhcp_dns' function tests: 1


def test_set_dhcp_dns(ethernet_config):
    """
    Test if it set DNS source to DHCP on Windows.
    """
    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.set_dhcp_dns("Ethernet") == {
            "DNS Server": "DHCP",
            "Interface": "Ethernet",
        }


# 'set_dhcp_all' function tests: 1


def test_set_dhcp_all(ethernet_config):
    """
    Test if it set both IP Address and DNS to DHCP.
    """
    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.set_dhcp_all("Ethernet") == {
            "Interface": "Ethernet",
            "DNS Server": "DHCP",
            "DHCP enabled": "Yes",
        }


# 'get_default_gateway' function tests: 1


def test_get_default_gateway(ethernet_config):
    """
    Test if it set DNS source to DHCP on Windows.
    """
    mock_cmd = MagicMock(return_value=ethernet_config)
    with patch.dict(win_ip.__salt__, {"cmd.run": mock_cmd}):
        assert win_ip.get_default_gateway() == "1.2.3.1"



import pytest

import salt.states.mac_keychain as keychain
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {keychain: {}}


def test_install_cert():
    """
    Test installing a certificate into the macOS keychain
    """
    expected = {
        "changes": {"installed": "Friendly Name"},
        "comment": "",
        "name": "/path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(return_value=["Cert1"])
    friendly_mock = MagicMock(return_value="Friendly Name")
    install_mock = MagicMock(return_value="1 identity imported.")
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.install": install_mock,
        },
    ):
        out = keychain.installed("/path/to/cert.p12", "passw0rd")
        list_mock.assert_called_once_with("/Library/Keychains/System.keychain")
        friendly_mock.assert_called_once_with("/path/to/cert.p12", "passw0rd")
        install_mock.assert_called_once_with(
            "/path/to/cert.p12", "passw0rd", "/Library/Keychains/System.keychain"
        )
        assert out == expected


def test_installed_cert():
    """
    Test installing a certificate into the macOS keychain when it's
    already installed
    """
    expected = {
        "changes": {},
        "comment": "Friendly Name already installed.",
        "name": "/path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(return_value=["Friendly Name"])
    friendly_mock = MagicMock(return_value="Friendly Name")
    install_mock = MagicMock(return_value="1 identity imported.")
    hash_mock = MagicMock(return_value="ABCD")
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.install": install_mock,
            "keychain.get_hash": hash_mock,
        },
    ):
        out = keychain.installed("/path/to/cert.p12", "passw0rd")
        list_mock.assert_called_once_with("/Library/Keychains/System.keychain")
        friendly_mock.assert_called_once_with("/path/to/cert.p12", "passw0rd")
        assert not install_mock.called
        assert out == expected


def test_uninstall_cert():
    """
    Test uninstalling a certificate into the macOS keychain when it's
    already installed
    """
    expected = {
        "changes": {"uninstalled": "Friendly Name"},
        "comment": "",
        "name": "/path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(return_value=["Friendly Name"])
    friendly_mock = MagicMock(return_value="Friendly Name")
    uninstall_mock = MagicMock(return_value="1 identity imported.")
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.uninstall": uninstall_mock,
        },
    ):
        out = keychain.uninstalled("/path/to/cert.p12", "passw0rd")
        list_mock.assert_called_once_with("/Library/Keychains/System.keychain")
        friendly_mock.assert_called_once_with("/path/to/cert.p12", "passw0rd")
        uninstall_mock.assert_called_once_with(
            "Friendly Name", "/Library/Keychains/System.keychain", None
        )
        assert out == expected


def test_uninstalled_cert():
    """
    Test uninstalling a certificate into the macOS keychain when it's
    not installed
    """
    expected = {
        "changes": {},
        "comment": "Friendly Name already uninstalled.",
        "name": "/path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(return_value=["Cert2"])
    friendly_mock = MagicMock(return_value="Friendly Name")
    uninstall_mock = MagicMock(return_value="1 identity imported.")
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.uninstall": uninstall_mock,
        },
    ):
        out = keychain.uninstalled("/path/to/cert.p12", "passw0rd")
        list_mock.assert_called_once_with("/Library/Keychains/System.keychain")
        friendly_mock.assert_called_once_with("/path/to/cert.p12", "passw0rd")
        assert not uninstall_mock.called
        assert out == expected


def test_default_keychain():
    """
    Test setting the default keychain
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {"default": "/path/to/chain.keychain"},
            "comment": "",
            "name": "/path/to/chain.keychain",
            "result": True,
        }

        exists_mock.return_value = True
        get_default_mock = MagicMock(return_value="/path/to/other.keychain")
        set_mock = MagicMock(return_value="")
        with patch.dict(
            keychain.__salt__,
            {
                "keychain.get_default_keychain": get_default_mock,
                "keychain.set_default_keychain": set_mock,
            },
        ):
            out = keychain.default_keychain(
                "/path/to/chain.keychain", "system", "frank"
            )
            get_default_mock.assert_called_once_with("frank", "system")
            set_mock.assert_called_once_with(
                "/path/to/chain.keychain", "system", "frank"
            )
            assert out == expected


def test_default_keychain_set_already():
    """
    Test setting the default keychain when it's already set
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "/path/to/chain.keychain was already the default keychain.",
            "name": "/path/to/chain.keychain",
            "result": True,
        }

        exists_mock.return_value = True
        get_default_mock = MagicMock(return_value="/path/to/chain.keychain")
        set_mock = MagicMock(return_value="")
        with patch.dict(
            keychain.__salt__,
            {
                "keychain.get_default_keychain": get_default_mock,
                "keychain.set_default_keychain": set_mock,
            },
        ):
            out = keychain.default_keychain(
                "/path/to/chain.keychain", "system", "frank"
            )
            get_default_mock.assert_called_once_with("frank", "system")
            assert not set_mock.called
            assert out == expected


def test_default_keychain_missing():
    """
    Test setting the default keychain when the keychain is missing
    """
    with patch("os.path.exists") as exists_mock:
        expected = {
            "changes": {},
            "comment": "Keychain not found at /path/to/cert.p12",
            "name": "/path/to/cert.p12",
            "result": False,
        }

        exists_mock.return_value = False
        out = keychain.default_keychain("/path/to/cert.p12", "system", "frank")
        assert out == expected


def test_install_cert_salt_fileserver():
    """
    Test installing a certificate into the macOS keychain from the salt
    fileserver
    """
    expected = {
        "changes": {"installed": "Friendly Name"},
        "comment": "",
        "name": "salt://path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(return_value=["Cert1"])
    friendly_mock = MagicMock(return_value="Friendly Name")
    install_mock = MagicMock(return_value="1 identity imported.")
    cp_cache_mock = MagicMock(return_value="/tmp/path/to/cert.p12")
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.install": install_mock,
            "cp.cache_file": cp_cache_mock,
        },
    ):
        out = keychain.installed("salt://path/to/cert.p12", "passw0rd")
        list_mock.assert_called_once_with("/Library/Keychains/System.keychain")
        friendly_mock.assert_called_once_with("/tmp/path/to/cert.p12", "passw0rd")
        install_mock.assert_called_once_with(
            "/tmp/path/to/cert.p12",
            "passw0rd",
            "/Library/Keychains/System.keychain",
        )
        assert out == expected


def test_installed_cert_hash_different():
    """
    Test installing a certificate into the macOS keychain when it's
    already installed but the certificate has changed
    """
    expected = {
        "changes": {"installed": "Friendly Name", "uninstalled": "Friendly Name"},
        "comment": (
            "Found a certificate with the same name but different hash, removing it.\n"
        ),
        "name": "/path/to/cert.p12",
        "result": True,
    }

    list_mock = MagicMock(side_effect=[["Friendly Name"], []])
    friendly_mock = MagicMock(return_value="Friendly Name")
    install_mock = MagicMock(return_value="1 identity imported.")
    uninstall_mock = MagicMock(return_value="removed.")
    hash_mock = MagicMock(side_effect=["ABCD", "XYZ"])
    with patch.dict(
        keychain.__salt__,
        {
            "keychain.list_certs": list_mock,
            "keychain.get_friendly_name": friendly_mock,
            "keychain.install": install_mock,
            "keychain.uninstall": uninstall_mock,
            "keychain.get_hash": hash_mock,
        },
    ):
        out = keychain.installed("/path/to/cert.p12", "passw0rd")
        list_mock.assert_has_calls(
            calls=[
                call("/Library/Keychains/System.keychain"),
                call("/Library/Keychains/System.keychain"),
            ]
        )
        friendly_mock.assert_called_once_with("/path/to/cert.p12", "passw0rd")
        install_mock.assert_called_once_with(
            "/path/to/cert.p12", "passw0rd", "/Library/Keychains/System.keychain"
        )
        uninstall_mock.assert_called_once_with(
            "Friendly Name",
            "/Library/Keychains/System.keychain",
            keychain_password=None,
        )
        assert out == expected



import getpass
import logging
import os

import pytest

import salt.modules.file as filemod
import salt.utils.files
import salt.utils.platform
from tests.support.mock import Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filemod: {
            "__context__": {},
            "__opts__": {"test": False},
        }
    }


@pytest.fixture
def tfile(tmp_path):
    filename = str(tmp_path / "file-check-test-file")

    with salt.utils.files.fopen(filename, "w") as fp:
        fp.write("Hi hello! I am a file.")
    os.chmod(filename, 0o644)

    yield filename

    os.remove(filename)


@pytest.fixture
def a_link(tmp_path, tfile):
    linkname = str(tmp_path / "a_link")
    os.symlink(tfile, linkname)

    yield linkname

    os.remove(linkname)


def get_link_perms():
    if salt.utils.platform.is_linux():
        return "0777"
    return "0755"


@pytest.mark.skip_on_windows(reason="os.symlink is not available on Windows")
def test_check_file_meta_follow_symlinks(a_link, tfile):
    user = getpass.getuser()
    lperms = get_link_perms()

    # follow_symlinks=False (default)
    ret = filemod.check_file_meta(
        a_link, tfile, None, None, user, None, lperms, None, None
    )
    assert ret == {}

    ret = filemod.check_file_meta(
        a_link, tfile, None, None, user, None, "0644", None, None
    )
    assert ret == {"mode": "0644"}

    # follow_symlinks=True
    ret = filemod.check_file_meta(
        a_link, tfile, None, None, user, None, "0644", None, None, follow_symlinks=True
    )
    assert ret == {}


@pytest.mark.skip_on_windows(reason="os.symlink is not available on Windows")
def test_check_managed_follow_symlinks(a_link, tfile):
    user = getpass.getuser()
    lperms = get_link_perms()

    # Function check_managed() ignores mode changes for files in the temp directory.
    # Trick it to not recognize a_link as such.
    a_link = "/" + a_link

    # follow_symlinks=False (default)
    ret, comments = filemod.check_managed(
        a_link, tfile, None, None, user, None, lperms, None, None, None, None, None
    )
    assert ret is True
    assert comments == f"The file {a_link} is in the correct state"

    ret, comments = filemod.check_managed(
        a_link, tfile, None, None, user, None, "0644", None, None, None, None, None
    )
    assert ret is None
    assert comments == "The following values are set to be changed:\nmode: 0644\n"

    # follow_symlinks=True
    ret, comments = filemod.check_managed(
        a_link,
        tfile,
        None,
        None,
        user,
        None,
        "0644",
        None,
        None,
        None,
        None,
        None,
        follow_symlinks=True,
    )
    assert ret is True
    assert comments == f"The file {a_link} is in the correct state"


@pytest.mark.skip_on_windows(reason="os.symlink is not available on Windows")
def test_check_managed_changes_follow_symlinks(a_link, tfile):
    user = getpass.getuser()
    lperms = get_link_perms()

    # follow_symlinks=False (default)
    ret = filemod.check_managed_changes(
        a_link, tfile, None, None, user, None, lperms, None, None, None, None, None
    )
    assert ret == {}

    ret = filemod.check_managed_changes(
        a_link, tfile, None, None, user, None, "0644", None, None, None, None, None
    )
    assert ret == {"mode": "0644"}

    # follow_symlinks=True
    ret = filemod.check_managed_changes(
        a_link,
        tfile,
        None,
        None,
        user,
        None,
        "0644",
        None,
        None,
        None,
        None,
        None,
        follow_symlinks=True,
    )
    assert ret == {}


@pytest.mark.skip_on_windows(reason="os.symlink is not available on Windows")
@pytest.mark.parametrize(
    "input,expected",
    [
        # user/group changes needed by name
        (
            {"user": "cuser", "group": "cgroup"},
            {"user": "cuser", "group": "cgroup"},
        ),
        # no changes needed by name
        (
            {"user": "luser", "group": "lgroup"},
            {},
        ),
        # user/group changes needed by id
        (
            {"user": 1001, "group": 2001},
            {"user": 1001, "group": 2001},
        ),
        # no user/group changes needed by id
        (
            {"user": 3001, "group": 4001},
            {},
        ),
    ],
)
def test_check_perms_user_group_name_and_id(input, expected):
    filename = "/path/to/fnord"
    with patch("os.path.exists", Mock(return_value=True)):
        # Consistent initial file stats
        stat_out = {
            "user": "luser",
            "group": "lgroup",
            "uid": 3001,
            "gid": 4001,
            "mode": "123",
        }

        patch_stats = patch(
            "salt.modules.file.stats",
            Mock(return_value=stat_out),
        )

        # "chown" the file to the permissions we want in test["input"]
        # pylint: disable=W0640
        def fake_chown(cmd, *args, **kwargs):
            for k, v in input.items():
                stat_out.update({k: v})

        patch_chown = patch(
            "salt.modules.file.chown",
            Mock(side_effect=fake_chown),
        )

        with patch_stats, patch_chown:
            ret, pre_post = filemod.check_perms(
                name=filename,
                ret={},
                user=input["user"],
                group=input["group"],
                mode="123",
                follow_symlinks=False,
            )
            assert ret["changes"] == expected



"""
Infoblox host record management.

functions accept api_opts:

    api_verifyssl: verify SSL [default to True or pillar value]
    api_url: server to connect to [default to pillar value]
    api_username:  [default to pillar value]
    api_password:  [default to pillar value]
"""


def present(name=None, start_addr=None, end_addr=None, data=None, **api_opts):
    """
    Ensure range record is present.

    infoblox_range.present:
        start_addr: '129.97.150.160',
        end_addr: '129.97.150.170',

    Verbose state example:

    .. code-block:: yaml

        infoblox_range.present:
            data: {
                'always_update_dns': False,
                'authority': False,
                'comment': 'range of IP addresses used for salt.. was used for ghost images deployment',
                'ddns_generate_hostname': True,
                'deny_all_clients': False,
                'deny_bootp': False,
                'disable': False,
                'email_list': [],
                'enable_ddns': False,
                'enable_dhcp_thresholds': False,
                'enable_email_warnings': False,
                'enable_ifmap_publishing': False,
                'enable_snmp_warnings': False,
                'end_addr': '129.97.150.169',
                'exclude': [],
                'extattrs': {},
                'fingerprint_filter_rules': [],
                'high_water_mark': 95,
                'high_water_mark_reset': 85,
                'ignore_dhcp_option_list_request': False,
                'lease_scavenge_time': -1,
                'logic_filter_rules': [],
                'low_water_mark': 0,
                'low_water_mark_reset': 10,
                'mac_filter_rules': [],
                'member': {'_struct': 'dhcpmember',
                        'ipv4addr': '129.97.128.9',
                        'name': 'cn-dhcp-mc.example.ca'},
                'ms_options': [],
                'nac_filter_rules': [],
                'name': 'ghost-range',
                'network': '129.97.150.0/24',
                'network_view': 'default',
                'option_filter_rules': [],
                'options': [{'name': 'dhcp-lease-time',
                            'num': 51,
                            'use_option': False,
                            'value': '43200',
                            'vendor_class': 'DHCP'}],
                'recycle_leases': True,
                'relay_agent_filter_rules': [],
                'server_association_type': 'MEMBER',
                'start_addr': '129.97.150.160',
                'update_dns_on_lease_renewal': False,
                'use_authority': False,
                'use_bootfile': False,
                'use_bootserver': False,
                'use_ddns_domainname': False,
                'use_ddns_generate_hostname': True,
                'use_deny_bootp': False,
                'use_email_list': False,
                'use_enable_ddns': False,
                'use_enable_dhcp_thresholds': False,
                'use_enable_ifmap_publishing': False,
                'use_ignore_dhcp_option_list_request': False,
                'use_known_clients': False,
                'use_lease_scavenge_time': False,
                'use_nextserver': False,
                'use_options': False,
                'use_recycle_leases': False,
                'use_unknown_clients': False,
                'use_update_dns_on_lease_renewal': False
            }
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not data:
        data = {}
    if "name" not in data:
        data.update({"name": name})
    if "start_addr" not in data:
        data.update({"start_addr": start_addr})
    if "end_addr" not in data:
        data.update({"end_addr": end_addr})

    obj = __salt__["infoblox.get_ipv4_range"](
        data["start_addr"], data["end_addr"], **api_opts
    )
    if obj is None:
        obj = __salt__["infoblox.get_ipv4_range"](
            start_addr=data["start_addr"], end_addr=None, **api_opts
        )
        if obj is None:
            obj = __salt__["infoblox.get_ipv4_range"](
                start_addr=None, end_addr=data["end_addr"], **api_opts
            )

    if obj:
        diff = __salt__["infoblox.diff_objects"](data, obj)
        if not diff:
            ret["result"] = True
            ret["comment"] = "supplied fields in correct state"
            return ret
        if diff:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "would attempt to update record"
                return ret
            new_obj = __salt__["infoblox.update_object"](
                obj["_ref"], data=data, **api_opts
            )
            ret["result"] = True
            ret["comment"] = "record fields updated"
            ret["changes"] = {"diff": diff}
            return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"would attempt to create record {name}"
        return ret

    new_obj_ref = __salt__["infoblox.create_ipv4_range"](data, **api_opts)
    new_obj = __salt__["infoblox.get_ipv4_range"](
        data["start_addr"], data["end_addr"], **api_opts
    )

    ret["result"] = True
    ret["comment"] = "record created"
    ret["changes"] = {"old": "None", "new": {"_ref": new_obj_ref, "data": new_obj}}
    return ret


def absent(name=None, start_addr=None, end_addr=None, data=None, **api_opts):
    """
    Ensure the range is removed

    Supplying the end of the range is optional.

    State example:

    .. code-block:: yaml

        infoblox_range.absent:
            - name: 'vlan10'

        infoblox_range.absent:
            - name:
            - start_addr: 127.0.1.20
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not data:
        data = {}
    if "name" not in data:
        data.update({"name": name})
    if "start_addr" not in data:
        data.update({"start_addr": start_addr})
    if "end_addr" not in data:
        data.update({"end_addr": end_addr})

    obj = __salt__["infoblox.get_ipv4_range"](
        data["start_addr"], data["end_addr"], **api_opts
    )
    if obj is None:
        obj = __salt__["infoblox.get_ipv4_range"](
            start_addr=data["start_addr"], end_addr=None, **api_opts
        )
        if obj is None:
            obj = __salt__["infoblox.get_ipv4_range"](
                start_addr=None, end_addr=data["end_addr"], **api_opts
            )

    if not obj:
        ret["result"] = True
        ret["comment"] = "already deleted"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "would attempt to delete range"
        return ret

    if __salt__["infoblox.delete_object"](objref=obj["_ref"]):
        ret["result"] = True
        ret["changes"] = {
            "old": f"Found {start_addr} - {end_addr}",
            "new": "Removed",
        }
    return ret



import pytest

import salt.utils.json


@pytest.mark.slow_test
@pytest.mark.netapi_client_data(["local", "runner"])
async def test_all_jobs(http_client, auth_creds, content_type_map):
    """
    test query to /jobs returns job data
    """
    # Login to get the token
    body = salt.utils.json.dumps(auth_creds)
    response = await http_client.fetch(
        "/login",
        method="POST",
        body=body,
        headers={
            "Accept": content_type_map["json"],
            "Content-Type": content_type_map["json"],
        },
    )
    assert response.code == 200
    token = response.headers["X-Auth-Token"]

    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}
    body = salt.utils.json.dumps(low)
    # Add a job
    response = await http_client.fetch(
        "/run",
        method="POST",
        body=body,
        headers={
            "Accept": content_type_map["json"],
            "Content-Type": content_type_map["json"],
        },
    )
    assert response.code == 200
    body = salt.utils.json.loads(response.body)

    # Get Jobs
    response = await http_client.fetch(
        "/jobs",
        method="GET",
        headers={"Accept": content_type_map["json"], "X-Auth-Token": token},
    )
    assert response.code == 200
    body = salt.utils.json.loads(response.body)
    for ret in body["return"][0].values():
        assert "Function" in ret
        if ret["Function"] == "test.ping":
            break
    else:
        pytest.fail("Failed to get the 'test.ping' job")



"""
A general map/reduce style salt runner for aggregating results
returned by several different minions.

.. versionadded:: 2014.7.0

Aggregated results are sorted by the size of the minion pools which returned
matching results.

Useful for playing the game: *"some of these things are not like the others..."*
when identifying discrepancies in a large infrastructure managed by salt.
"""

import salt.client
from salt.exceptions import SaltClientError


def hash(*args, **kwargs):
    """
    Return the MATCHING minion pools from the aggregated and sorted results of
    a salt command

    .. versionadded:: 2014.7.0

    This command is submitted via a salt runner using the
    general form::

        salt-run survey.hash [survey_sort=up/down] <target>
                  <salt-execution-module> <salt-execution-module parameters>

    Optionally accept a ``survey_sort=`` parameter. Default: ``survey_sort=down``

    CLI Example #1: (functionally equivalent to ``salt-run manage.up``)

    .. code-block:: bash

        salt-run survey.hash "*" test.ping

    CLI Example #2: (find an "outlier" minion config file)

    .. code-block:: bash

        salt-run survey.hash "*" file.get_hash /etc/salt/minion survey_sort=up
    """

    return _get_pool_results(*args, **kwargs)


def diff(*args, **kwargs):
    """
    Return the DIFFERENCE of the result sets returned by each matching minion
    pool

    .. versionadded:: 2014.7.0

    These pools are determined from the aggregated and sorted results of
    a salt command.

    This command displays the "diffs" as a series of 2-way differences --
    namely the difference between the FIRST displayed minion pool
    (according to sort order) and EACH SUBSEQUENT minion pool result set.

    Differences are displayed according to the Python ``difflib.unified_diff()``
    as in the case of the salt execution module ``file.get_diff``.

    This command is submitted via a salt runner using the general form::

        salt-run survey.diff [survey_sort=up/down] <target>
                     <salt-execution-module> <salt-execution-module parameters>

    Optionally accept a ``survey_sort=`` parameter. Default:
    ``survey_sort=down``

    CLI Example #1: (Example to display the "differences of files")

    .. code-block:: bash

        salt-run survey.diff survey_sort=up "*" cp.get_file_str file:///etc/hosts
    """
    # TODO: The salt execution module "cp.get_file_str file:///..." is a
    # non-obvious way to display the differences between files using
    # survey.diff .  A more obvious method needs to be found or developed.

    import difflib

    bulk_ret = _get_pool_results(*args, **kwargs)

    is_first_time = True
    for k in bulk_ret:
        print("minion pool :\n------------")
        print(k["pool"])
        print("pool size :\n----------")
        print("    " + str(len(k["pool"])))
        if is_first_time:
            is_first_time = False
            print("pool result :\n------------")
            print("    " + bulk_ret[0]["result"])
            print()
            continue

        outs = 'differences from "{}" results :'.format(bulk_ret[0]["pool"][0])
        print(outs)
        print("-" * (len(outs) - 1))
        from_result = bulk_ret[0]["result"].splitlines()
        for idx, _ in enumerate(from_result):
            from_result[idx] += "\n"
        to_result = k["result"].splitlines()
        for idx, _ in enumerate(to_result):
            to_result[idx] += "\n"
        outs = ""
        outs += "".join(
            difflib.unified_diff(
                from_result,
                to_result,
                fromfile=bulk_ret[0]["pool"][0],
                tofile=k["pool"][0],
                n=0,
            )
        )
        print(outs)
        print()

    return bulk_ret


def _get_pool_results(*args, **kwargs):
    """
    A helper function which returns a dictionary of minion pools along with
    their matching result sets.
    Useful for developing other "survey style" functions.
    Optionally accepts a "survey_sort=up" or "survey_sort=down" kwargs for
    specifying sort order.
    Because the kwargs namespace of the "salt" and "survey" command are shared,
    the name "survey_sort" was chosen to help avoid option conflicts.
    """
    # TODO: the option "survey.sort=" would be preferred for namespace
    # separation but the kwargs parser for the salt-run command seems to
    # improperly pass the options containing a "." in them for later modules to
    # process. The "_" is used here instead.

    import hashlib

    tgt = args[0]
    cmd = args[1]
    ret = {}

    sort = kwargs.pop("survey_sort", "down")
    direction = sort != "up"

    tgt_type = kwargs.pop("tgt_type", "compound")
    if tgt_type not in ["compound", "pcre"]:
        tgt_type = "compound"

    kwargs_passthru = {
        key: value for (key, value) in kwargs.items() if not key.startswith("_")
    }

    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            minions = client.cmd(
                tgt,
                cmd,
                args[2:],
                timeout=__opts__["timeout"],
                tgt_type=tgt_type,
                kwarg=kwargs_passthru,
            )
        except SaltClientError as client_error:
            print(client_error)
            return ret

    # hash minion return values as a string
    for minion in sorted(minions):
        digest = hashlib.sha256(
            str(minions[minion]).encode(__salt_system_encoding__)
        ).hexdigest()
        if digest not in ret:
            ret[digest] = {}
            ret[digest]["pool"] = []
            ret[digest]["result"] = str(minions[minion])

        ret[digest]["pool"].append(minion)

    sorted_ret = []
    for k in sorted(ret, key=lambda k: len(ret[k]["pool"]), reverse=direction):
        # return aggregated results, sorted by size of the hash pool

        sorted_ret.append(ret[k])

    return sorted_ret



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.launchctl
"""

import pytest

import salt.modules.launchctl_service as launchctl
import salt.utils.stringutils
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {launchctl: {}}


def test_get_all():
    """
    Test for Return all installed services
    """
    with patch.dict(
        launchctl.__salt__, {"cmd.run": MagicMock(return_value="A\tB\tC\t\n")}
    ):
        with patch.object(
            launchctl, "_available_services", return_value={"A": "a", "B": "b"}
        ):
            assert launchctl.get_all() == ["A", "B", "C"]


def test_available():
    """
    Test for Check that the given service is available.
    """
    with patch.object(launchctl, "_service_by_name", return_value=True):
        assert launchctl.available("job_label")


def test_missing():
    """
    Test for The inverse of service.available
    """
    with patch.object(launchctl, "_service_by_name", return_value=True):
        assert not launchctl.missing("job_label")


def test_status():
    """
    Test for Return the status for a service
    """
    launchctl_data = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>Label</key>
    <string>salt-minion</string>
    <key>LastExitStatus</key>
    <integer>0</integer>
    <key>LimitLoadToSessionType</key>
    <string>System</string>
    <key>OnDemand</key>
    <false/>
    <key>PID</key>
    <integer>71</integer>
    <key>ProgramArguments</key>
    <array>
        <string>/usr/local/bin/salt-minion</string>
    </array>
    <key>TimeOut</key>
    <integer>30</integer>
</dict>
</plist>"""
    with patch.object(
        launchctl, "_service_by_name", return_value={"plist": {"Label": "A"}}
    ):
        launchctl_data = salt.utils.stringutils.to_bytes(launchctl_data)
        with patch.object(
            launchctl, "_get_launchctl_data", return_value=launchctl_data
        ):
            assert launchctl.status("job_label")


def test_stop():
    """
    Test for Stop the specified service
    """
    with patch.object(launchctl, "_service_by_name", return_value={"file_path": "A"}):
        with patch.dict(
            launchctl.__salt__, {"cmd.retcode": MagicMock(return_value=False)}
        ):
            assert launchctl.stop("job_label")

    with patch.object(launchctl, "_service_by_name", return_value=None):
        assert not launchctl.stop("job_label")


def test_start():
    """
    Test for Start the specified service
    """
    with patch.object(launchctl, "_service_by_name", return_value={"file_path": "A"}):
        with patch.dict(
            launchctl.__salt__, {"cmd.retcode": MagicMock(return_value=False)}
        ):
            assert launchctl.start("job_label")

    with patch.object(launchctl, "_service_by_name", return_value=None):
        assert not launchctl.start("job_label")


def test_restart():
    """
    Test for Restart the named service
    """
    with patch.object(launchctl, "stop", return_value=None):
        with patch.object(launchctl, "start", return_value=True):
            assert launchctl.restart("job_label")



"""
General management functions for salt, tools like seeing what hosts are up
and what hosts are down
"""

import logging
import operator
import os
import re
import subprocess
import tempfile
import time
import urllib.request
import uuid

import salt.client
import salt.client.ssh
import salt.key
import salt.utils.compat
import salt.utils.files
import salt.utils.minions
import salt.utils.path
import salt.utils.versions
import salt.version
import salt.wheel
from salt.exceptions import SaltClientError, SaltSystemExit

FINGERPRINT_REGEX = re.compile(r"^([a-f0-9]{2}:){15}([a-f0-9]{2})$")

log = logging.getLogger(__name__)


def _ping(tgt, tgt_type, timeout, gather_job_timeout):
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        pub_data = client.run_job(
            tgt, "test.ping", (), tgt_type, "", timeout, "", listen=True
        )

        if not pub_data:
            return pub_data

        log.debug(
            "manage runner will ping the following minion(s): %s",
            ", ".join(sorted(pub_data["minions"])),
        )

        returned = set()
        for fn_ret in client.get_cli_event_returns(
            pub_data["jid"],
            pub_data["minions"],
            client._get_timeout(timeout),
            tgt,
            tgt_type,
            gather_job_timeout=gather_job_timeout,
        ):

            if fn_ret:
                for mid, _ in fn_ret.items():
                    log.debug("minion '%s' returned from ping", mid)
                    returned.add(mid)

        not_returned = sorted(set(pub_data["minions"]) - returned)
        returned = sorted(returned)

        return returned, not_returned


def status(
    output=True, tgt="*", tgt_type="glob", timeout=None, gather_job_timeout=None
):
    """
    .. versionchanged:: 2017.7.0

        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Print the status of all known salt minions

    CLI Example:

    .. code-block:: bash

        salt-run manage.status
        salt-run manage.status tgt="webservers" tgt_type="nodegroup"
        salt-run manage.status timeout=5 gather_job_timeout=10
    """
    ret = {}

    if not timeout:
        timeout = __opts__["timeout"]
    if not gather_job_timeout:
        gather_job_timeout = __opts__["gather_job_timeout"]

    res = _ping(tgt, tgt_type, timeout, gather_job_timeout)
    ret["up"], ret["down"] = ([], []) if not res else res
    return ret


def key_regen():
    """
    This routine is used to regenerate all keys in an environment. This is
    invasive! ALL KEYS IN THE SALT ENVIRONMENT WILL BE REGENERATED!!

    The key_regen routine sends a command out to minions to revoke the master
    key and remove all minion keys, it then removes all keys from the master
    and prompts the user to restart the master. The minions will all reconnect
    and keys will be placed in pending.

    After the master is restarted and minion keys are in the pending directory
    execute a salt-key -A command to accept the regenerated minion keys.

    The master *must* be restarted within 60 seconds of running this command or
    the minions will think there is something wrong with the keys and abort.

    Only Execute this runner after upgrading minions and master to 0.15.1 or
    higher!

    CLI Example:

    .. code-block:: bash

        salt-run manage.key_regen
    """
    client = salt.client.get_local_client(__opts__["conf_file"])
    try:
        client.cmd("*", "saltutil.regen_keys")
    except SaltClientError as client_error:
        print(client_error)
        return False

    for root, _, files in salt.utils.path.os_walk(__opts__["pki_dir"]):
        for fn_ in files:
            path = os.path.join(root, fn_)
            try:
                os.remove(path)
            except OSError:
                pass
    msg = (
        "The minion and master keys have been deleted.  Restart the Salt\n"
        "Master within the next 60 seconds!!!\n\n"
        "Wait for the minions to reconnect.  Once the minions reconnect\n"
        "the new keys will appear in pending and will need to be re-\n"
        "accepted by running:\n"
        "    salt-key -A\n\n"
        "Be advised that minions not currently connected to the master\n"
        "will not be able to reconnect and may require manual\n"
        "regeneration via a local call to\n"
        "    salt-call saltutil.regen_keys"
    )
    return msg


def down(
    removekeys=False, tgt="*", tgt_type="glob", timeout=None, gather_job_timeout=None
):
    """
    .. versionchanged:: 2017.7.0

        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Print a list of all the down or unresponsive salt minions
    Optionally remove keys of down minions

    CLI Example:

    .. code-block:: bash

        salt-run manage.down
        salt-run manage.down removekeys=True
        salt-run manage.down tgt="webservers" tgt_type="nodegroup"
    """
    ret = status(
        output=False,
        tgt=tgt,
        tgt_type=tgt_type,
        timeout=timeout,
        gather_job_timeout=gather_job_timeout,
    ).get("down", [])
    for minion in ret:
        if removekeys:
            wheel = salt.wheel.Wheel(__opts__)
            wheel.call_func("key.delete", match=minion)
    return ret


def up(
    tgt="*", tgt_type="glob", timeout=None, gather_job_timeout=None
):  # pylint: disable=C0103
    """
    .. versionchanged:: 2017.7.0

        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Print a list of all of the minions that are up

    CLI Example:

    .. code-block:: bash

        salt-run manage.up
        salt-run manage.up tgt="webservers" tgt_type="nodegroup"
        salt-run manage.up timeout=5 gather_job_timeout=10
    """
    ret = status(
        output=False,
        tgt=tgt,
        tgt_type=tgt_type,
        timeout=timeout,
        gather_job_timeout=gather_job_timeout,
    ).get("up", [])
    return ret


def list_state(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.list_state
    """
    # Always return 'present' for 0MQ for now
    # TODO: implement other states support for 0MQ
    ckminions = salt.utils.minions.CkMinions(__opts__)
    minions = ckminions.connected_ids(show_ip=show_ip, subset=subset)

    connected = dict(minions) if show_ip else sorted(minions)

    return connected


def list_not_state(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.list_not_state
    """
    connected = list_state(subset=None, show_ip=show_ip)

    with salt.key.get_key(__opts__) as key:
        keys = key.list_keys()

        not_connected = []
        for minion in keys[key.ACC]:
            if minion not in connected and (subset is None or minion in subset):
                not_connected.append(minion)

        return not_connected


def present(subset=None, show_ip=False):
    """
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.present
    """
    return list_state(subset=subset, show_ip=show_ip)


def not_present(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.5.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.not_present
    """
    return list_not_state(subset=subset, show_ip=show_ip)


def joined(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.joined
    """
    return list_state(subset=subset, show_ip=show_ip)


def not_joined(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.not_joined
    """
    return list_not_state(subset=subset, show_ip=show_ip)


def allowed(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.allowed
    """
    return list_state(subset=subset, show_ip=show_ip)


def not_allowed(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.not_allowed
    """
    return list_not_state(subset=subset, show_ip=show_ip)


def alived(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.alived
    """
    return list_state(subset=subset, show_ip=show_ip)


def not_alived(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.not_alived
    """
    return list_not_state(subset=subset, show_ip=show_ip)


def reaped(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are up according to Salt's presence
    detection (no commands will be sent to minions)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.reaped
    """
    return list_state(subset=subset, show_ip=show_ip)


def not_reaped(subset=None, show_ip=False):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2019.2.0

    Print a list of all minions that are NOT up according to Salt's presence
    detection (no commands will be sent)

    subset : None
        Pass in a list of minion ids.

    show_ip : False
        Also show the IP address each minion is connecting from.

    CLI Example:

    .. code-block:: bash

        salt-run manage.not_reaped
    """
    return list_not_state(subset=subset, show_ip=show_ip)


def safe_accept(target, tgt_type="glob"):
    """
    .. versionchanged:: 2017.7.0

        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Accept a minion's public key after checking the fingerprint over salt-ssh

    CLI Example:

    .. code-block:: bash

        salt-run manage.safe_accept my_minion
        salt-run manage.safe_accept minion1,minion2 tgt_type=list
    """
    ssh_client = salt.client.ssh.client.SSHClient()
    ret = ssh_client.cmd(target, "key.finger", tgt_type=tgt_type)

    failures = {}
    for minion, finger in ret.items():
        if not FINGERPRINT_REGEX.match(finger):
            failures[minion] = finger
        else:
            with salt.key.Key(__opts__) as salt_key:
                fingerprints = salt_key.finger(minion)
            accepted = fingerprints.get("minions", {})
            pending = fingerprints.get("minions_pre", {})
            if minion in accepted:
                del ret[minion]
                continue
            elif minion not in pending:
                failures[minion] = f"Minion key {minion} not found by salt-key"
            elif pending[minion] != finger:
                failures[minion] = (
                    "Minion key {} does not match the key in salt-key: {}".format(
                        finger, pending[minion]
                    )
                )
            else:
                subprocess.call(["salt-key", "-qya", minion])

        if minion in failures:
            del ret[minion]

    if failures:
        print("safe_accept failed on the following minions:")
        for minion, message in failures.items():
            print(minion)
            print("-" * len(minion))
            print(message)
            print("")

    __jid_event__.fire_event({"message": f"Accepted {len(ret):d} keys"}, "progress")
    return ret, failures


def versions():
    """
    Check the version of active minions

    CLI Example:

    .. code-block:: bash

        salt-run manage.versions
    """
    ret = {}
    client = salt.client.get_local_client(__opts__["conf_file"])
    try:
        minions = client.cmd(
            "*", "test.version", full_return=True, timeout=__opts__["timeout"]
        )
    except SaltClientError as client_error:
        print(client_error)
        return ret

    labels = {
        -2: "Minion offline",
        -1: "Minion requires update",
        0: "Up to date",
        1: "Minion newer than master",
        2: "Master",
    }

    version_status = {}

    master_version = salt.version.__saltstack_version__

    for minion in minions:
        if not minions[minion] or minions[minion]["retcode"]:
            minion_version = False
            ver_diff = -2
        else:
            minion_version = salt.version.SaltStackVersion.parse(minions[minion]["ret"])
            ver_diff = salt.utils.compat.cmp(minion_version, master_version)

        if ver_diff not in version_status:
            version_status[ver_diff] = {}
        if minion_version:
            version_status[ver_diff][minion] = minion_version.string
        else:
            version_status[ver_diff][minion] = minion_version

    # Add version of Master to output
    version_status[2] = master_version.string

    for key in version_status:
        if key == 2:
            ret[labels[key]] = version_status[2]
        else:
            for minion in sorted(version_status[key]):
                ret.setdefault(labels[key], {})[minion] = version_status[key][minion]
    return ret


def bootstrap(
    version="develop",
    script="https://github.com/saltstack/salt-bootstrap/releases/latest/download/bootstrap-salt.sh",
    hosts="",
    script_args="",
    roster="flat",
    ssh_user=None,
    ssh_password=None,
    ssh_priv_key=None,
    tmp_dir="/tmp/.bootstrap",
    http_backend="tornado",
):
    """
    Bootstrap minions with salt-bootstrap

    version : develop
        Git tag of version to install

    script : https://github.com/saltstack/salt-bootstrap/releases/latest/download/bootstrap-salt.sh
        URL containing the script to execute

    hosts
        Comma-separated hosts [example: hosts='host1.local,host2.local']. These
        hosts need to exist in the specified roster.

    script_args
        Any additional arguments that you want to pass to the script.

        .. versionadded:: 2016.11.0

    roster : flat
        The roster to use for Salt SSH. More information about roster files can
        be found in :ref:`Salt's Roster Documentation <ssh-roster>`.

        A full list of roster types, see the :ref:`builtin roster modules <all-salt.roster>`
        documentation.

        .. versionadded:: 2016.11.0

    ssh_user
        If ``user`` isn't found in the ``roster``, a default SSH user can be set here.
        Keep in mind that ``ssh_user`` will not override the roster ``user`` value if
        it is already defined.

        .. versionadded:: 2016.11.0

    ssh_password
        If ``passwd`` isn't found in the ``roster``, a default SSH password can be set
        here. Keep in mind that ``ssh_password`` will not override the roster ``passwd``
        value if it is already defined.

        .. versionadded:: 2016.11.0

    ssh_privkey
        If ``priv`` isn't found in the ``roster``, a default SSH private key can be set
        here. Keep in mind that ``ssh_password`` will not override the roster ``passwd``
        value if it is already defined.

        .. versionadded:: 2016.11.0

    tmp_dir : /tmp/.bootstrap
        The temporary directory to download the bootstrap script in. This
        directory will have ``-<uuid4>`` appended to it. For example:
        ``/tmp/.bootstrap-a19a728e-d40a-4801-aba9-d00655c143a7/``

        .. versionadded:: 2016.11.0

    http_backend : tornado
        The backend library to use to download the script. If you need to use
        a ``file:///`` URL, then you should set this to ``urllib2``.

        .. versionadded:: 2016.11.0

    CLI Example:

    .. code-block:: bash

        salt-run manage.bootstrap hosts='host1,host2'
        salt-run manage.bootstrap hosts='host1,host2' version='v3006.2'
        salt-run manage.bootstrap hosts='host1,host2' version='v3006.2' script='https://github.com/saltstack/salt-bootstrap/develop'
    """

    client_opts = __opts__.copy()
    if roster is not None:
        client_opts["roster"] = roster

    if ssh_user is not None:
        client_opts["ssh_user"] = ssh_user

    if ssh_password is not None:
        client_opts["ssh_passwd"] = ssh_password

    if ssh_priv_key is not None:
        client_opts["ssh_priv"] = ssh_priv_key

    for host in hosts.split(","):
        client_opts["tgt"] = host
        client_opts["selected_target_option"] = "glob"
        tmp_dir = "{}-{}/".format(tmp_dir.rstrip("/"), uuid.uuid4())
        deploy_command = os.path.join(tmp_dir, "deploy.sh")
        try:
            client_opts["argv"] = ["file.makedirs", tmp_dir, "mode=0700"]
            salt.client.ssh.SSH(client_opts).run()
            client_opts["argv"] = [
                "http.query",
                script,
                f"backend={http_backend}",
                f"text_out={deploy_command}",
            ]
            salt.client.ssh.SSH(client_opts).run()
            client_opts["argv"] = [
                "cmd.run",
                " ".join(["sh", deploy_command, script_args]),
                "python_shell=False",
            ]
            salt.client.ssh.SSH(client_opts).run()
            client_opts["argv"] = ["file.remove", tmp_dir]
            salt.client.ssh.SSH(client_opts).run()
        except SaltSystemExit as exc:
            log.error(str(exc))


def bootstrap_psexec(
    hosts="",
    master=None,
    version=None,
    arch="win32",
    installer_url=None,
    username=None,
    password=None,
):
    """
    Bootstrap Windows minions via PsExec.

    hosts
        Comma separated list of hosts to deploy the Windows Salt minion.

    master
        Address of the Salt master passed as an argument to the installer.

    version
        Point release of installer to download. Defaults to the most recent.

    arch
        Architecture of installer to download. Defaults to win32.

    installer_url
        URL of minion installer executable. Defaults to the latest version from
        https://packages.broadcom.com/artifactory/saltproject-generic/windows/

    username
        Optional user name for login on remote computer.

    password
        Password for optional username. If omitted, PsExec will prompt for one
        to be entered for each host.

    CLI Example:

    .. code-block:: bash

        salt-run manage.bootstrap_psexec hosts='host1,host2'
        salt-run manage.bootstrap_psexec hosts='host1,host2' version='0.17' username='DOMAIN\\Administrator'
        salt-run manage.bootstrap_psexec hosts='host1,host2' installer_url='http://exampledomain/salt-installer.exe'
    """

    # TODO: Need to make this gets the latest version from the new repo location
    # TODO: Similar to tests/support/win_installer.py
    # TODO: Maybe need to move that ^^^^ to a salt util
    if not installer_url:
        base_url = "https://repo.saltproject.io/windows/"
        source = urllib.request.urlopen(base_url).read()
        salty_rx = re.compile(
            '>(Salt-Minion-(.+?)-(.+)-Setup.exe)</a></td><td align="right">(.*?)\\s*<'
        )
        source_list = sorted(
            (
                [path, ver, plat, time.strptime(date, "%d-%b-%Y %H:%M")]
                for path, ver, plat, date in salty_rx.findall(source)
            ),
            key=operator.itemgetter(3),
            reverse=True,
        )
        if version:
            source_list = [s for s in source_list if s[1] == version]
        if arch:
            source_list = [s for s in source_list if s[2] == arch]

        if not source_list:
            return -1

        version = source_list[0][1]
        arch = source_list[0][2]
        installer_url = base_url + source_list[0][0]

    # It's no secret that Windows is notoriously command-line hostile.
    # Win 7 and newer can use PowerShell out of the box, but to reach
    # all those XP and 2K3 machines we must suppress our gag-reflex
    # and use VB!

    # The following script was borrowed from an informative article about
    # downloading exploit payloads for malware. Nope, no irony here.
    # http://www.greyhathacker.net/?p=500
    vb_script = """strFileURL = "{0}"
strHDLocation = "{1}"
Set objXMLHTTP = CreateObject("MSXML2.XMLHTTP")
objXMLHTTP.open "GET", strFileURL, false
objXMLHTTP.send()
If objXMLHTTP.Status = 200 Then
Set objADOStream = CreateObject("ADODB.Stream")
objADOStream.Open
objADOStream.Type = 1
objADOStream.Write objXMLHTTP.ResponseBody
objADOStream.Position = 0
objADOStream.SaveToFile strHDLocation
objADOStream.Close
Set objADOStream = Nothing
End if
Set objXMLHTTP = Nothing
Set objShell = CreateObject("WScript.Shell")
objShell.Exec("{1}{2}")"""

    vb_saltexec = "saltinstall.exe"
    vb_saltexec_args = " /S /minion-name=%COMPUTERNAME%"
    if master:
        vb_saltexec_args += f" /master={master}"

    # One further thing we need to do; the Windows Salt minion is pretty
    # self-contained, except for the Microsoft Visual C++ 2008 runtime.
    # It's tiny, so the bootstrap will attempt a silent install.
    vb_vcrunexec = "vcredist.exe"
    if arch == "AMD64":
        vb_vcrun = vb_script.format(
            "http://download.microsoft.com/download/d/2/4/d242c3fb-da5a-4542-ad66-f9661d0a8d19/vcredist_x64.exe",
            vb_vcrunexec,
            " /q",
        )
    else:
        vb_vcrun = vb_script.format(
            "http://download.microsoft.com/download/d/d/9/dd9a82d0-52ef-40db-8dab-795376989c03/vcredist_x86.exe",
            vb_vcrunexec,
            " /q",
        )

    vb_salt = vb_script.format(installer_url, vb_saltexec, vb_saltexec_args)

    # PsExec doesn't like extra long arguments; save the instructions as a batch
    # file so we can fire it over for execution.

    # First off, change to the local temp directory, stop salt-minion (if
    # running), and remove the master's public key.
    # This is to accommodate for reinstalling Salt over an old or broken build,
    # e.g. if the master address is changed, the salt-minion process will fail
    # to authenticate and quit; which means infinite restarts under Windows.
    batch = (
        "cd /d %TEMP%\nnet stop salt-minion\ndel"
        " c:\\salt\\conf\\pki\\minion\\minion_master.pub\n"
    )

    # Speaking of command-line hostile, cscript only supports reading a script
    # from a file. Glue it together line by line.
    for x, y in ((vb_vcrunexec, vb_vcrun), (vb_saltexec, vb_salt)):
        vb_lines = y.split("\n")
        batch += (
            "\ndel "
            + x
            + "\n@echo "
            + vb_lines[0]
            + "  >"
            + x
            + ".vbs\n@echo "
            + ("  >>" + x + ".vbs\n@echo ").join(vb_lines[1:])
            + "  >>"
            + x
            + ".vbs\ncscript.exe /NoLogo "
            + x
            + ".vbs"
        )

    batch_path = tempfile.mkstemp(suffix=".bat")[1]
    with salt.utils.files.fopen(batch_path, "wb") as batch_file:
        batch_file.write(batch)

    for host in hosts.split(","):
        argv = ["psexec", "\\\\" + host]
        if username:
            argv += ["-u", username]
            if password:
                argv += ["-p", password]
        argv += ["-h", "-c", batch_path]
        subprocess.call(argv)



"""
Package support for the dummy proxy used by the test suite
"""

import logging

import salt.utils.data
import salt.utils.platform

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pkg"


def __virtual__():
    """
    Only work on systems that are a proxy minion
    """
    try:
        if salt.utils.platform.is_proxy() and __opts__["proxy"]["proxytype"] == "dummy":
            return __virtualname__
    except KeyError:
        return (
            False,
            "The dummyproxy_package execution module failed to load. Check "
            "the proxy key in pillar or /etc/salt/proxy.",
        )

    return (
        False,
        "The dummyproxy_package execution module failed to load: only works "
        "on a dummy proxy minion.",
    )


def list_pkgs(versions_as_list=False, **kwargs):
    return __proxy__["dummy.package_list"]()


def install(name=None, refresh=False, fromrepo=None, pkgs=None, sources=None, **kwargs):
    return __proxy__["dummy.package_install"](name, **kwargs)


def remove(name=None, pkgs=None, **kwargs):
    return __proxy__["dummy.package_remove"](name)


def version(*names, **kwargs):
    """
    Returns a string representing the package version or an empty string if not
    installed. If more than one package name is specified, a dict of
    name/version pairs is returned.

    CLI Example:

    .. code-block:: bash

        salt '*' pkg.version <package name>
        salt '*' pkg.version <package1> <package2> <package3> ...
    """
    if len(names) == 1:
        vers = __proxy__["dummy.package_status"](names[0])
        return vers[names[0]]
    else:
        results = {}
        for n in names:
            vers = __proxy__["dummy.package_status"](n)
            results.update(vers)
        return results


def upgrade(
    name=None, pkgs=None, refresh=True, skip_verify=True, normalize=True, **kwargs
):
    old = __proxy__["dummy.package_list"]()
    new = __proxy__["dummy.uptodate"]()
    pkg_installed = __proxy__["dummy.upgrade"]()
    ret = salt.utils.data.compare_dicts(old, pkg_installed)
    return ret


def installed(
    name,
    version=None,
    refresh=False,
    fromrepo=None,
    skip_verify=False,
    pkgs=None,
    sources=None,
    **kwargs
):

    p = __proxy__["dummy.package_status"](name)
    if version is None:
        if "ret" in p:
            return str(p["ret"])
        else:
            return True
    else:
        if p is not None:
            return version == str(p)



"""
    :codeauthor: Gareth J. Greenaway <gareth@saltstack.com>

    Test cases for salt.states.netconfig
"""

import pytest

import salt.modules.napalm_network as net_mod
import salt.states.netconfig as netconfig
import salt.utils.files
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    state_loader_globals = {
        "__env__": "base",
        "__salt__": {"net.replace_pattern": net_mod.replace_pattern},
    }
    module_loader_globals = {
        "__env__": "base",
        "__salt__": {
            "net.replace_pattern": net_mod.replace_pattern,
            "net.load_config": net_mod.load_config,
        },
    }
    return {netconfig: state_loader_globals, net_mod: module_loader_globals}


def test_replace_pattern_test_is_true():
    """
    Test to replace_pattern to ensure that test=True
    is being passed correctly.
    """
    name = "name"
    pattern = "OLD-POLICY-NAME"
    repl = "new-policy-name"

    mock = MagicMock()
    mock_net_replace_pattern = MagicMock()
    mock_loaded_ret = MagicMock()

    with patch.dict(netconfig.__salt__, {"config.merge": mock}):
        with patch.dict(
            netconfig.__salt__, {"net.replace_pattern": mock_net_replace_pattern}
        ):
            with patch.object(salt.utils.napalm, "loaded_ret", mock_loaded_ret):
                # Test if test=True is passed as argument to state.apply
                with patch.dict(netconfig.__opts__, {"test": True}):
                    netconfig.replace_pattern(name, pattern, repl)

                    # Get the args and kwargs from the mocked call net.replace_pattern
                    args, kwargs = mock_net_replace_pattern.call_args_list[0]

                    # Verify that the keyword argument is True
                    assert kwargs["test"]

                    # Get the args and kwargs from the mocked call to salt.utils.napalm.loaded_ret
                    args, kwargs = mock_loaded_ret.call_args_list[0]

                    # Verify that the third positional argument is True
                    assert args[2]

                # Test if test=True is passed as argument to state directly
                netconfig.replace_pattern(name, pattern, repl, test=True)

                # Get the args and kwargs from the mocked call net.replace_pattern
                args, kwargs = mock_net_replace_pattern.call_args_list[0]

                # Verify that the keyword argument is True
                assert kwargs["test"]

                # Get the args and kwargs from the mocked call to salt.utils.napalm.loaded_ret
                args, kwargs = mock_loaded_ret.call_args_list[0]

                # Verify that the third positional argument is True
                assert args[2]


def test_managed_test_is_true():
    """
    Test to managed to ensure that test=True
    is being passed correctly.
    """
    name = "name"

    mock = MagicMock()
    mock_update_config = MagicMock()

    with patch.dict(netconfig.__salt__, {"config.merge": mock}):
        with patch.object(netconfig, "_update_config", mock_update_config):
            # Test if test=True is passed as argument to state.apply
            with patch.dict(netconfig.__opts__, {"test": True}):
                netconfig.managed(name)

                # Get the args and kwargs from the mocked call net.replace_pattern
                args, kwargs = mock_update_config.call_args_list[0]

                # Verify that the keyword argument is True
                assert kwargs["test"]

            # Test if test=True is passed as argument to state directly
            netconfig.managed(name, test=True)

            # Get the args and kwargs from the mocked call net.replace_pattern
            args, kwargs = mock_update_config.call_args_list[0]

            # Verify that the keyword argument is True
            assert kwargs["test"]



"""
Manage Perl modules using CPAN

.. versionadded:: 2015.5.0
"""

import logging
import os
import os.path

import salt.utils.files
import salt.utils.path

log = logging.getLogger(__name__)

# Don't shadow built-ins.
__func_alias__ = {"list_": "list"}


def __virtual__():
    """
    Only work on supported POSIX-like systems
    """
    if salt.utils.path.which("cpan"):
        return True
    return (False, "Unable to locate cpan. Make sure it is installed and in the PATH.")


def install(module):
    """
    Install a Perl module from CPAN

    CLI Example:

    .. code-block:: bash

        salt '*' cpan.install Template::Alloy
    """
    ret = {
        "old": None,
        "new": None,
    }

    old_info = show(module)

    cmd = f"cpan -i {module}"
    out = __salt__["cmd.run"](cmd)

    if "don't know what it is" in out:
        ret["error"] = "CPAN cannot identify this package"
        return ret

    new_info = show(module)
    ret["old"] = old_info.get("installed version", None)
    ret["new"] = new_info["installed version"]

    return ret


def remove(module, details=False):
    """
    Attempt to remove a Perl module that was installed from CPAN. Because the
    ``cpan`` command doesn't actually support "uninstall"-like functionality,
    this function will attempt to do what it can, with what it has from CPAN.

    Until this function is declared stable, USE AT YOUR OWN RISK!

    CLI Example:

    .. code-block:: bash

        salt '*' cpan.remove Old::Package
    """
    ret = {
        "old": None,
        "new": None,
    }

    info = show(module)
    if "error" in info:
        return {"error": info["error"]}

    version = info.get("installed version", None)
    if version is None:
        return ret

    ret["old"] = version

    if "cpan build dirs" not in info:
        return {"error": "No CPAN data available to use for uninstalling"}

    mod_pathfile = module.replace("::", "/") + ".pm"
    ins_path = info["installed file"].replace(mod_pathfile, "")

    files = []
    for build_dir in info["cpan build dirs"]:
        contents = os.listdir(build_dir)
        if "MANIFEST" not in contents:
            continue
        mfile = os.path.join(build_dir, "MANIFEST")
        with salt.utils.files.fopen(mfile, "r") as fh_:
            for line in fh_.readlines():
                line = salt.utils.stringutils.to_unicode(line)
                if line.startswith("lib/"):
                    files.append(line.replace("lib/", ins_path).strip())

    rm_details = {}
    for file_ in files:
        if file_ in rm_details:
            continue
        log.trace("Removing %s", file_)
        if __salt__["file.remove"](file_):
            rm_details[file_] = "removed"
        else:
            rm_details[file_] = "unable to remove"

    if details:
        ret["details"] = rm_details

    return ret


def list_():
    """
    List installed Perl modules, and the version installed

    CLI Example:

    .. code-block:: bash

        salt '*' cpan.list
    """
    ret = {}
    cmd = "cpan -l"
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        comps = line.split()
        ret[comps[0]] = comps[1]
    return ret


def show(module):
    """
    Show information about a specific Perl module

    CLI Example:

    .. code-block:: bash

        salt '*' cpan.show Template::Alloy
    """
    ret = {}
    ret["name"] = module

    # This section parses out details from CPAN, if possible
    cmd = f"cpan -D {module}"
    out = __salt__["cmd.run"](cmd).splitlines()
    mode = "skip"
    info = []
    for line in out:
        if line.startswith("-------------"):
            mode = "parse"
            continue
        if mode == "skip":
            continue
        info.append(line)

    if len(info) == 6:
        # If the module is not installed, we'll be short a line
        info.insert(2, "")
    if len(info) < 6:
        # This must not be a real package
        ret["error"] = "This package does not seem to exist"
        return ret

    ret["description"] = info[0].strip()
    ret["cpan file"] = info[1].strip()
    if info[2].strip():
        ret["installed file"] = info[2].strip()
    else:
        ret["installed file"] = None
    comps = info[3].split(":")
    if len(comps) > 1:
        ret["installed version"] = comps[1].strip()
    if "installed version" not in ret or not ret["installed version"]:
        ret["installed version"] = None
    comps = info[4].split(":")
    comps = comps[1].split()
    ret["cpan version"] = comps[0].strip()
    ret["author name"] = info[5].strip()
    ret["author email"] = info[6].strip()

    # Check and see if there are cpan build directories
    config = show_config()
    build_dir = config.get("build_dir", None)
    if build_dir is not None:
        ret["cpan build dirs"] = []
        builds = os.listdir(build_dir)
        pfile = module.replace("::", "-")
        for file_ in builds:
            if file_.startswith(pfile):
                ret["cpan build dirs"].append(os.path.join(build_dir, file_))

    return ret


def show_config():
    """
    Return a dict of CPAN configuration values

    CLI Example:

    .. code-block:: bash

        salt '*' cpan.show_config
    """
    ret = {}
    cmd = "cpan -J"
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        if "=>" not in line:
            # TODO: Some options take up multiple lines, so this doesn't always work
            continue
        comps = line.split("=>")
        key = comps[0].replace("'", "").strip()
        val = comps[1].replace("',", "").replace("'", "").strip()
        ret[key] = val
    return ret



"""
Tests the SaltfileMixIn.
"""

import optparse  # pylint: disable=deprecated-module
import shutil

import pytest

import salt.exceptions
import salt.utils.parsers
from tests.support.helpers import patched_environ
from tests.support.mock import patch


class MockSaltfileParser(
    salt.utils.parsers.OptionParser,
    salt.utils.parsers.SaltfileMixIn,
    metaclass=salt.utils.parsers.OptionParserMeta,
):
    def __init__(self, *args, **kwargs):
        salt.utils.parsers.OptionParser.__init__(self, *args, **kwargs)
        self.config = {}

    def _mixin_setup(self):
        self.add_option(
            "-l",
            "--log-level",
            dest="log_level",
            default="warning",
            help="The log level for salt.",
        )
        group = self.output_options_group = optparse.OptionGroup(
            self, "Output Options", "Configure your preferred output format."
        )
        self.add_option_group(group)

        group.add_option(
            "--out",
            "--output",
            dest="output",
            help=(
                "Print the output from the '{}' command using the "
                "specified outputter.".format(
                    self.get_prog_name(),
                )
            ),
        )
        group.add_option(
            "--out-file",
            "--output-file",
            dest="output_file",
            default=None,
            help="Write the output to the specified file.",
        )
        group.add_option(
            "--version-arg",
            action="version",
            help="Option to test no dest",
        )


@pytest.fixture
def parser():
    return MockSaltfileParser()


@pytest.fixture
def saltfile(tmp_path):
    fp = tmp_path / "Saltfile"
    fp.touch()
    return fp


def test_saltfile_in_environment(parser, saltfile):
    """
    Test setting the SALT_SALTFILE environment variable
    """
    with patched_environ(SALT_SALTFILE=str(saltfile)):
        parser.parse_args([])
        assert parser.options.saltfile == str(saltfile)


def test_saltfile_option(parser, saltfile):
    """
    Test setting the saltfile via the CLI
    """
    parser.parse_args(["--saltfile", str(saltfile)])
    assert parser.options.saltfile == str(saltfile)


def test_bad_saltfile_option(parser, saltfile, tmp_path):
    """
    Test setting a bad saltfile via the CLI
    """
    with pytest.raises(SystemExit):
        parser.parse_args(["--saltfile", str(tmp_path / "fake_dir")])


def test_saltfile_cwd(parser, saltfile, tmp_path):
    """
    Test using a saltfile in the cwd
    """
    with patch("os.getcwd", return_value=str(tmp_path)) as cwd_mock:
        parser.parse_args([])
        assert parser.options.saltfile == str(saltfile)
        cwd_mock.assert_called_once()


def test_saltfile_cwd_doesnt_exist(parser, saltfile, tmp_path):
    """
    Test using a saltfile in the cwd that doesn't exist
    """
    with patch("os.getcwd", return_value=str(tmp_path / "fake_dir")) as cwd_mock:
        parser.parse_args([])
        assert parser.options.saltfile is None


def test_saltfile_user_home(parser, saltfile, tmp_path):
    """
    Test using a saltfile in ~/.salt/
    """
    fake_dir = tmp_path / "fake_dir"
    fake_dir.mkdir()
    with patch("os.getcwd", return_value=str(fake_dir)) as cwd_mock:
        with patch("os.path.expanduser", return_value=str(tmp_path)) as eu_mock:
            salt_subdir = tmp_path / ".salt"
            salt_subdir.mkdir()
            dest = str(salt_subdir / "Saltfile")
            shutil.copy(str(saltfile), dest)
            parser.parse_args([])
            assert parser.options.saltfile == dest
            cwd_mock.assert_called_once()
            eu_mock.assert_called_with("~")


def test_bad_saltfile(parser, saltfile):
    """
    Test a saltfile with bad configuration
    """
    contents = """
    bad "yaml":
  - this is: bad yaml
  -       bad yaml=data:
    - {"bad": yaml, "data": "yaml"}
    """
    saltfile.write_text(contents)
    # It raises two errors, let's catch them both
    with pytest.raises(SystemExit):
        with pytest.raises(salt.exceptions.SaltConfigurationError):
            parser.parse_args(["--saltfile", str(saltfile)])


def test_saltfile_without_prog_name(parser, saltfile):
    """
    Test a saltfile with valid yaml but without the program name in it
    """
    contents = "good: yaml"
    saltfile.write_text(contents)
    # This should just run cleanly
    parser.parse_args(["--saltfile", str(saltfile)])


def test_saltfile(parser, saltfile):
    """
    Test a valid saltfile
    """
    contents = """
    __main__.py:
      log_level: debug
      output: json
    """
    saltfile.write_text(contents)
    parser.parse_args(["--saltfile", str(saltfile)])
    print(parser.option_list)
    assert parser.options.log_level == "debug"
    assert parser.options.output == "json"


def test_saltfile_unusual_option(parser, saltfile):
    """
    Test a valid saltfile
    """
    contents = """
    __main__.py:
      go: birds
    """
    saltfile.write_text(contents)
    parser.parse_args(["--saltfile", str(saltfile)])
    assert parser.options.go == "birds"


def test_saltfile_cli_override(parser, saltfile):
    """
    Test a valid saltfile
    """
    contents = """
    __main__.py:
      log_level: debug
      output: json
      output_file: /fake/file
    """
    saltfile.write_text(contents)
    parser.parse_args(
        [
            "--saltfile",
            str(saltfile),
            "--log-level",
            "info",
            "--out-file",
            "/still/fake/file",
        ]
    )
    assert parser.options.log_level == "info"
    assert parser.options.output == "json"
    assert parser.options.output_file == "/still/fake/file"



"""
Support for freebsd-update utility on FreeBSD.

.. versionadded:: 2017.7.0

:maintainer:    George Mamalakis <mamalos@gmail.com>
:maturity:      new
:platform:      FreeBSD
"""

import logging

import salt.utils.path
from salt.exceptions import CommandNotFoundError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "freebsd_update"
__virtual_aliases__ = ("freebsd-update",)


def __virtual__():
    """
    .. versionadded:: 2016.3.4

    Only work on FreeBSD RELEASEs >= 6.2, where freebsd-update was introduced.
    """
    if __grains__["os"] != "FreeBSD":
        return (
            False,
            "The freebsd_update execution module cannot be loaded: only available on"
            " FreeBSD systems.",
        )
    if float(__grains__["osrelease"]) < 6.2:
        return (
            False,
            "freebsd_update is only available on FreeBSD versions >= 6.2-RELESE",
        )
    if "release" not in __grains__["kernelrelease"].lower():
        return (False, "freebsd_update is only available on FreeBSD RELEASES")
    return __virtualname__


def _cmd(**kwargs):
    """
    .. versionadded:: 2016.3.4

    Private function that returns the freebsd-update command string to be
    executed. It checks if any arguments are given to freebsd-update and appends
    them accordingly.
    """
    update_cmd = salt.utils.path.which("freebsd-update")
    if not update_cmd:
        raise CommandNotFoundError('"freebsd-update" command not found')

    params = []
    if "basedir" in kwargs:
        params.append("-b {}".format(kwargs["basedir"]))
    if "workdir" in kwargs:
        params.append("-d {}".format(kwargs["workdir"]))
    if "conffile" in kwargs:
        params.append("-f {}".format(kwargs["conffile"]))
    if "force" in kwargs:
        params.append("-F")
    if "key" in kwargs:
        params.append("-k {}".format(kwargs["key"]))
    if "newrelease" in kwargs:
        params.append("-r {}".format(kwargs["newrelease"]))
    if "server" in kwargs:
        params.append("-s {}".format(kwargs["server"]))
    if "address" in kwargs:
        params.append("-t {}".format(kwargs["address"]))

    if params:
        return "{} {}".format(update_cmd, " ".join(params))
    return update_cmd


def _wrapper(orig, pre="", post="", err_=None, run_args=None, **kwargs):
    """
    Helper function that wraps the execution of freebsd-update command.

    orig:
        Originating function that called _wrapper().

    pre:
        String that will be prepended to freebsd-update command.

    post:
        String that will be appended to freebsd-update command.

    err_:
        Dictionary on which return codes and stout/stderr are copied.

    run_args:
        Arguments to be passed on cmd.run_all.

    kwargs:
        Parameters of freebsd-update command.
    """
    ret = ""  # the message to be returned
    cmd = _cmd(**kwargs)
    cmd_str = " ".join([x for x in (pre, cmd, post, orig)])
    if run_args and isinstance(run_args, dict):
        res = __salt__["cmd.run_all"](cmd_str, **run_args)
    else:
        res = __salt__["cmd.run_all"](cmd_str)

    if isinstance(err_, dict):  # copy return values if asked to
        for k, v in res.items():
            err_[k] = v

    if "retcode" in res and res["retcode"] != 0:
        msg = " ".join([x for x in (res["stdout"], res["stderr"]) if x])
        ret = 'Unable to run "{}" with run_args="{}". Error: {}'.format(
            cmd_str, run_args, msg
        )
        log.error(ret)
    else:
        try:
            ret = res["stdout"]
        except KeyError:
            log.error(
                "cmd.run_all did not return a dictionary with a key named 'stdout'"
            )
    return ret


def fetch(**kwargs):
    """
    .. versionadded:: 2016.3.4

    freebsd-update fetch wrapper. Based on the currently installed world and the
    configuration options set, fetch all available binary updates.

    kwargs:
        Parameters of freebsd-update command.
    """
    # fetch continues when no controlling terminal is present
    pre = ""
    post = ""
    run_args = {}
    if float(__grains__["osrelease"]) >= 10.2:
        post += "--not-running-from-cron"
    else:
        pre += " env PAGER=cat"
        run_args["python_shell"] = True
    return _wrapper("fetch", pre=pre, post=post, run_args=run_args, **kwargs)


def install(**kwargs):
    """
    .. versionadded:: 2016.3.4

    freebsd-update install wrapper. Install the most recently fetched updates or
    upgrade.

    kwargs:
        Parameters of freebsd-update command.
    """
    return _wrapper("install", **kwargs)


def rollback(**kwargs):
    """
    .. versionadded:: 2016.3.4

    freebsd-update rollback wrapper. Uninstalls the most recently installed
    updates.

    kwargs:
        Parameters of freebsd-update command.
    """
    return _wrapper("rollback", **kwargs)


def update(**kwargs):
    """
    .. versionadded:: 2016.3.4

    Command that simplifies freebsd-update by running freebsd-update fetch first
    and then freebsd-update install.

    kwargs:
        Parameters of freebsd-update command.
    """
    stdout = {}

    for mode in ("fetch", "install"):
        err_ = {}
        ret = _wrapper(mode, err_=err_, **kwargs)
        if "retcode" in err_ and err_["retcode"] != 0:
            return ret
        if "stdout" in err_:
            stdout[mode] = err_["stdout"]
    return "\n".join([f"{k}: {v}" for (k, v) in stdout.items()])


def ids(**kwargs):
    """
    .. versionadded:: 2016.3.4

    freebsd-update IDS wrapper function. Compares the system against a "known
    good" index of the installed release.

    kwargs:
        Parameters of freebsd-update command.
    """
    return _wrapper("IDS", **kwargs)


def upgrade(**kwargs):
    """
    .. versionadded:: 2016.3.4

    Dummy function used only to print a message that upgrade is not available.
    The reason is that upgrade needs manual intervention and reboot, so even if
    used with:

       yes | freebsd-upgrade -r VERSION

    the additional freebsd-update install that needs to run after the reboot
    cannot be implemented easily.

    kwargs:
        Parameters of freebsd-update command.
    """
    msg = "freebsd-update upgrade not yet implemented."
    log.warning(msg)
    return msg



"""
Use a git repository as a Pillar source
---------------------------------------

This external pillar allows for a Pillar top file and Pillar SLS files to be
sourced from a git repository.

However, since git_pillar does not have an equivalent to the
:conf_master:`pillar_roots` parameter, configuration is slightly different. A
Pillar top file is required to be in the git repository and must still contain
the relevant environment, like so:

.. code-block:: yaml

    base:
      '*':
        - foo

The branch/tag which maps to that environment must then be specified along with
the repo's URL. Configuration details can be found below.

.. important::
    Each branch/tag used for git_pillar must have its own top file. This is
    different from how the top file works when configuring :ref:`States
    <states-tutorial>`. The reason for this is that each git_pillar branch/tag
    is processed separately from the rest. Therefore, if the ``qa`` branch is
    to be used for git_pillar, it would need to have its own top file, with the
    ``qa`` environment defined within it, like this:

    .. code-block:: yaml

        qa:
          'dev-*':
            - bar

    Additionally, while git_pillar allows for the branch/tag to be overridden
    (see :ref:`here <git-pillar-env-remap>`), keep in
    mind that the top file must reference the actual environment name. It is
    common practice to make the environment in a git_pillar top file match the
    branch/tag name, but when remapping, the environment of course no longer
    matches the branch/tag, and the top file needs to be adjusted accordingly.
    When expected Pillar values configured in git_pillar are missing, this is a
    common misconfiguration that may be to blame, and is a good first step in
    troubleshooting.

.. _git-pillar-configuration:

Configuring git_pillar for Salt
===============================

Beginning with Salt version 2015.8.0, pygit2_ is now supported in addition to
GitPython_. The requirements for GitPython_ and pygit2_ are the same as for
GitFS, as described :ref:`here <gitfs-dependencies>`.

.. important::
    git_pillar has its own set of global configuration parameters. While it may
    seem intuitive to use the global gitfs configuration parameters
    (:conf_master:`gitfs_base`, etc.) to manage git_pillar, this will not work.
    The main difference for this is the fact that the different components
    which use Salt's git backend code do not all function identically. For
    instance, in git_pillar it is necessary to specify which branch/tag to be
    used for git_pillar remotes. This is the reverse behavior from gitfs, where
    branches/tags make up your environments.

    See :ref:`here <git-pillar-config-opts>` for documentation on the
    git_pillar configuration options and their usage.

Here is an example git_pillar configuration:

.. code-block:: yaml

    ext_pillar:
      - git:
        # Use 'prod' instead of the branch name 'production' as the environment
        - production https://gitserver/git-pillar.git:
          - env: prod
        # Use 'dev' instead of the branch name 'develop' as the environment
        - develop https://gitserver/git-pillar.git:
          - env: dev
        # No per-remote config parameters (and no trailing colon), 'qa' will
        # be used as the environment
        - qa https://gitserver/git-pillar.git
        # SSH key authentication
        - master git@other-git-server:pillardata-ssh.git:
          # Pillar SLS files will be read from the 'pillar' subdirectory in
          # this repository
          - root: pillar
          - privkey: /path/to/key
          - pubkey: /path/to/key.pub
          - passphrase: CorrectHorseBatteryStaple
        # HTTPS authentication
        - master https://other-git-server/pillardata-https.git:
          - user: git
          - password: CorrectHorseBatteryStaple

The main difference between this and the old way of configuring git_pillar is
that multiple remotes can be configured under one ``git`` section under
:conf_master:`ext_pillar`. More than one ``git`` section can be used, but it is
not necessary. Remotes will be evaluated sequentially.

Per-remote configuration parameters are supported (similar to :ref:`gitfs
<gitfs-per-remote-config>`), and global versions of the git_pillar
configuration parameters can also be set.

.. _git-pillar-env-remap:

To remap a specific branch to a specific Pillar environment, use the ``env``
per-remote parameter:

.. code-block:: yaml

    ext_pillar:
      - git:
        - production https://gitserver/git-pillar.git:
          - env: prod

If ``__env__`` is specified as the branch name, then git_pillar will decide
which branch to use based on the following criteria:

- If the minion has a :conf_minion:`pillarenv` configured, it will use that
  pillar environment. (2016.11.2 and later)
- Otherwise, if the minion has an ``environment`` configured, it will use that
  environment.
- Otherwise, the master's :conf_master:`git_pillar_base` will be used.

.. note::
    The use of :conf_minion:`environment` to choose the pillar environment
    dates from a time before the :conf_minion:`pillarenv` parameter was added.
    In a future release, it will be ignored and either the minion's
    :conf_minion:`pillarenv` or the master's :conf_master:`git_pillar_base`
    will be used.

Here's an example of using ``__env__`` as the git_pillar environment:

.. code-block:: yaml

    ext_pillar:
      - git:
        - __env__ https://gitserver/git-pillar.git:
          - root: pillar

The corresponding Pillar top file would look like this:

.. code-block:: yaml

    "{{saltenv}}":
      '*':
        - bar

With the addition of pygit2_ support, git_pillar can now interact with
authenticated remotes. Authentication works just like in gitfs (as outlined in
the :ref:`Git Fileserver Backend Walkthrough <gitfs-authentication>`), only
with the global authentication parameter names prefixed with ``git_pillar``
instead of ``gitfs`` (e.g. :conf_master:`git_pillar_pubkey`,
:conf_master:`git_pillar_privkey`, :conf_master:`git_pillar_passphrase`, etc.).

.. note::
    The ``name`` parameter can be used to further differentiate between two
    remotes with the same URL and branch. When using two remotes with the same
    URL, the ``name`` option is required.

.. _GitPython: https://github.com/gitpython-developers/GitPython
.. _pygit2: https://github.com/libgit2/pygit2

.. _git-pillar-multiple-remotes:

How Multiple Remotes Are Handled
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

As noted above, multiple remotes can be included in the same ``git`` ext_pillar
configuration. Consider the following:

.. code-block:: yaml

    my_etcd_config:
      etcd.host: 127.0.0.1
      etcd.port: 4001

    ext_pillar:
      - etcd: my_etcd_config
      - git:
        - master https://mydomain.tld/foo.git:
          - root: pillar
        - master https://mydomain.tld/bar.git
        - master https://mydomain.tld/baz.git
        - dev https://mydomain.tld/qux.git
      - git:
        - master https://mydomain.tld/abc.git
        - dev https://mydomain.tld/123.git

To understand how pillar data from these repos will be compiled, it's important
to know how Salt will process them. The following points should be kept in
mind:

1. Each ext_pillar is called separately from the others. So, in the above
   example, the :mod:`etcd <salt.pillar.etcd>` ext_pillar will be evaluated
   first, with the first group of git_pillar remotes evaluated next (and merged
   into the etcd pillar data). Lastly, the second group of git_pillar remotes
   will be evaluated, and then merged into the ext_pillar data evaluated before
   it.

2. Within a single group of git_pillar remotes, each remote will be evaluated in
   order, with results merged together as each remote is evaluated.

   .. note::
       Prior to the 2017.7.0 release, remotes would be evaluated in a
       non-deterministic order.

3. By default, when a repo is evaluated, other remotes' which share its pillar
   environment will have their files made available to the remote being
   processed.

The first point should be straightforward enough, but the second and third
could use some additional clarification.

First, point #2. In the first group of git_pillar remotes, the top file and
pillar SLS files in the ``foo`` remote will be evaluated first. The ``bar``
remote will be evaluated next, and its results will be merged into the pillar
data compiled when the ``foo`` remote was evaluated. As the subsequent remotes
are evaluated, their data will be merged in the same fashion.

But wait, don't these repositories belong to more than one pillar environments?
Well, yes. The default method of generating pillar data compiles pillar data
from all environments. This behavior can be overridden using a ``pillarenv``.
Setting a :conf_minion:`pillarenv` in the minion config file will make that
minion tell the master to ignore any pillar data from environments which don't
match that pillarenv. A pillarenv can also be specified for a given minion or
set of minions when :mod:`running states <salt.modules.state>`, by using the
``pillarenv`` argument. The CLI pillarenv will override one set in the minion
config file. So, assuming that a pillarenv of ``base`` was set for a minion, it
would not get any of the pillar variables configured in the ``qux`` remote,
since that remote is assigned to the ``dev`` environment. The only way to get
its pillar data would be to specify a pillarenv of ``dev``, which would mean
that it would then ignore any items from the ``base`` pillarenv. A more
detailed explanation of pillar environments can be found :ref:`here
<pillar-environments>`.

Moving on to point #3, and looking at the example ext_pillar configuration, as
the ``foo`` remote is evaluated, it will also have access to the files from the
``bar`` and ``baz`` remotes, since all three are assigned to the ``base``
pillar environment. So, if an SLS file referenced by the ``foo`` remotes's top
file does not exist in the ``foo`` remote, it will be searched for in the
``bar`` remote, followed by the ``baz`` remote. When it comes time to evaluate
the ``bar`` remote, SLS files referenced by the ``bar`` remote's top file will
first be looked for in the ``bar`` remote, followed by ``foo``, and ``baz``,
and when the ``baz`` remote is processed, SLS files will be looked for in
``baz``, followed by ``foo`` and ``bar``. This "failover" logic is called a
:ref:`directory overlay <file-roots-directory-overlay>`, and it is also used by
:conf_master:`file_roots` and :conf_minion`pillar_roots`. The ordering of which
remote is checked for SLS files is determined by the order they are listed.
First the remote being processed is checked, then the others that share the
same environment are checked. However, before the 2017.7.0 release, since
evaluation was unordered, the remote being processed would be checked, followed
in no specific order by the other repos which share the same environment.

Beginning with the 2017.7.0 release, this behavior of git_pillar remotes having
access to files in other repos which share the same environment can be disabled
by setting :conf_master:`git_pillar_includes` to ``False``. If this is done,
then all git_pillar remotes will only have access to their own SLS files.
Another way of ensuring that a git_pillar remote will not have access to SLS
files from other git_pillar remotes which share the same pillar environment is
to put them in a separate ``git`` section under ``ext_pillar``. Look again at
the example configuration above. In the second group of git_pillar remotes, the
``abc`` remote would not have access to the SLS files from the ``foo``,
``bar``, and ``baz`` remotes, and vice-versa.

.. _git-pillar-mountpoints:

Mountpoints
~~~~~~~~~~~

.. versionadded:: 2017.7.0

Assume the following pillar top file:

.. code-block:: yaml

    base:
      'web*':
        - common
        - web.server.nginx
        - web.server.appdata

Now, assume that you would like to configure the ``web.server.nginx`` and
``web.server.appdata`` SLS files in separate repos. This could be done using
the following ext_pillar configuration (assuming that
:conf_master:`git_pillar_includes` has not been set to ``False``):

.. code-block:: yaml

    ext_pillar:
      - git:
        - master https://mydomain.tld/pillar-common.git
        - master https://mydomain.tld/pillar-nginx.git
        - master https://mydomain.tld/pillar-appdata.git

However, in order to get the files in the second and third git_pillar remotes
to work, you would need to first create the directory structure underneath it
(i.e. place them underneath ``web/server/`` in the repository). This also makes
it tedious to reorganize the configuration, as changing ``web.server.nginx`` to
``web.nginx`` in the top file would require you to also move the SLS files in
the ``pillar-nginx`` up a directory level.

For these reasons, much like gitfs, git_pillar now supports a "mountpoint"
feature. Using the following ext_pillar configuration, the SLS files in the
second and third git_pillar remotes can be placed in the root of the git
repository:

.. code-block:: yaml

    ext_pillar:
      - git:
        - master https://mydomain.tld/pillar-common.git
        - master https://mydomain.tld/pillar-nginx.git:
          - mountpoint: web/server/
        - master https://mydomain.tld/pillar-appdata.git:
          - mountpoint: web/server/

Now, if the top file changed the SLS target from ``web.server.nginx``, instead
of reorganizing the git repository, you would just need to adjust the
mountpoint to ``web/`` (and restart the ``salt-master`` daemon).

.. note::
    - Leading and trailing slashes on the mountpoints are optional.
    - Use of the ``mountpoint`` feature requires that
      :conf_master:`git_pillar_includes` is not disabled.
    - Content from mounted git_pillar repos can only be referenced by a top
      file in the same pillar environment.
    - Salt versions prior to 2018.3.4 ignore the ``root`` parameter when
      ``mountpoint`` is set.

.. _git-pillar-all_saltenvs:

all_saltenvs
~~~~~~~~~~~~

.. versionadded:: 2018.3.4

When ``__env__`` is specified as the branch name, ``all_saltenvs`` per-remote configuration parameter overrides the logic Salt uses to map branches/tags to pillar environments (i.e. pillarenvs). This allows a single branch/tag to appear in all saltenvs. Example:

.. code-block:: yaml

    ext_pillar:
      - git:
        - __env__ https://mydomain.tld/top.git
          - all_saltenvs: master
        - __env__ https://mydomain.tld/pillar-nginx.git:
          - mountpoint: web/server/
        - __env__ https://mydomain.tld/pillar-appdata.git:
          - mountpoint: web/server/

.. _git_pillar_update_interval:

git_pillar_update_interval
~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 3000

This option defines the default update interval (in seconds) for git_pillar
remotes. The update is handled within the global loop, hence
``git_pillar_update_interval`` should be a multiple of ``loop_interval``.

.. code-block:: yaml

    git_pillar_update_interval: 120

.. _git-pillar-fallback:

fallback
~~~~~~~~

.. versionadded:: 3001

Setting ``fallback`` per-remote or global configuration parameter will map non-existing environments to a default branch. Example:

.. code-block:: yaml

    ext_pillar:
      - git:
        - __env__ https://mydomain.tld/top.git
          - all_saltenvs: master
        - __env__ https://mydomain.tld/pillar-nginx.git:
          - mountpoint: web/server/
          - fallback: master
        - __env__ https://mydomain.tld/pillar-appdata.git:
          - mountpoint: web/server/
          - fallback: master

"""

import copy
import logging

import salt.utils.dictupdate
import salt.utils.gitfs
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import FileserverConfigError
from salt.pillar import Pillar

PER_REMOTE_OVERRIDES = ("base", "env", "root", "ssl_verify", "refspecs", "fallback")
PER_REMOTE_ONLY = ("name", "mountpoint", "all_saltenvs")
GLOBAL_ONLY = ("branch",)

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "git"


def __virtual__():
    """
    Only load if gitpython is available
    """
    git_ext_pillars = [x for x in __opts__["ext_pillar"] if "git" in x]
    if not git_ext_pillars:
        # No git external pillars were configured
        return False

    try:
        salt.utils.gitfs.GitPillar(__opts__, init_remotes=False)
        # Initialization of the GitPillar object did not fail, so we
        # know we have valid configuration syntax and that a valid
        # provider was detected.
        return __virtualname__
    except FileserverConfigError:
        return False


def ext_pillar(minion_id, pillar, *repos):  # pylint: disable=unused-argument
    """
    Checkout the ext_pillar sources and compile the resulting pillar SLS
    """
    opts = copy.deepcopy(__opts__)
    opts["pillar_roots"] = {}
    opts["__git_pillar"] = True
    git_pillar = salt.utils.gitfs.GitPillar(
        opts,
        list(repos),
        per_remote_overrides=PER_REMOTE_OVERRIDES,
        per_remote_only=PER_REMOTE_ONLY,
        global_only=GLOBAL_ONLY,
    )
    if __opts__.get("__role") == "minion":
        # If masterless, fetch the remotes. We'll need to remove this once
        # we make the minion daemon able to run standalone.
        git_pillar.fetch_remotes()
    git_pillar.checkout()
    ret = {}
    merge_strategy = __opts__.get("pillar_source_merging_strategy", "smart")
    merge_lists = __opts__.get("pillar_merge_lists", False)
    for pillar_dir, env in git_pillar.pillar_dirs.items():
        # Map env if env == '__env__' before checking the env value
        if env == "__env__":
            env = (
                opts.get("pillarenv")
                or opts.get("saltenv")
                or opts.get("git_pillar_base")
            )
            log.debug("__env__ maps to %s", env)

        # If pillarenv is set, only grab pillars with that match pillarenv
        if opts["pillarenv"] and env != opts["pillarenv"]:
            log.debug(
                "env '%s' for pillar dir '%s' does not match pillarenv '%s', skipping",
                env,
                pillar_dir,
                opts["pillarenv"],
            )
            continue
        if pillar_dir in git_pillar.pillar_linked_dirs:
            log.debug(
                "git_pillar is skipping processing on %s as it is a mounted repo",
                pillar_dir,
            )
            continue
        else:
            log.debug(
                "git_pillar is processing pillar SLS from %s for pillar env '%s'",
                pillar_dir,
                env,
            )

        pillar_roots = [pillar_dir]

        if __opts__["git_pillar_includes"]:
            # Add the rest of the pillar_dirs in this environment to the
            # list, excluding the current pillar_dir being processed. This
            # is because it was already specified above as the first in the
            # list, so that its top file is sourced from the correct
            # location and not from another git_pillar remote.
            pillar_roots.extend(
                [
                    d
                    for (d, e) in git_pillar.pillar_dirs.items()
                    if env == e and d != pillar_dir
                ]
            )

        opts["pillar_roots"] = {env: pillar_roots}

        local_pillar = Pillar(opts, __grains__, minion_id, env)
        ret = salt.utils.dictupdate.merge(
            ret,
            local_pillar.compile_pillar(ext=False),
            strategy=merge_strategy,
            merge_lists=merge_lists,
        )
    return ret


def _extract_key_val(kv, delimiter="="):
    """Extract key and value from key=val string.

    Example:
    >>> _extract_key_val('foo=bar')
    ('foo', 'bar')
    """
    pieces = kv.split(delimiter)
    key = pieces[0]
    val = delimiter.join(pieces[1:])
    return key, val
